from __future__ import annotations

from pathlib import Path

import fire
import numpy as np
import pandas as pd
from omegaconf import DictConfig, ListConfig, OmegaConf

from carps.utils.loggingutils import get_logger, setup_logging

setup_logging()
logger = get_logger(__file__)


def add_scenario_type(logs: pd.DataFrame) -> pd.DataFrame:
    def determine_scenario_type(x: pd.Series) -> str:
        if x["task.is_multifidelity"] is False and x["task.is_multiobjective"] is False:
            scenario = "blackbox"
        elif x["task.is_multifidelity"] is True and x["task.is_multiobjective"] is False:
            scenario = "multi-fidelity"
        elif x["task.is_multifidelity"] is False and x["task.is_multiobjective"] is True:
            scenario = "multi-objective"
        elif x["task.is_multifidelity"] is True and x["task.is_multiobjective"] is True:
            scenario = "multi-fidelity-objective"
        elif np.isnan(x["task.is_multifidelity"]) or np.isnan(x["task.is_multiobjective"]):
            scenario = "blackbox"
        else:
            print(
                x["problem_id"],
                x["optimizer_id"],
                x["seed"],
                x["task.is_multifidelity"],
                type(x["task.is_multifidelity"]),
            )
            raise ValueError("Unknown scenario")
        return scenario

    logs["scenario"] = logs.apply(determine_scenario_type, axis=1)
    return logs


def maybe_postadd_task(logs: pd.DataFrame) -> pd.DataFrame:
    index_fn = Path(__file__).parent.parent / "configs/problem/index.csv"
    if not index_fn.is_file():
        raise ValueError("Problem ids have not been indexed. Run `python -m carps.utils.index_configs`.")
    problem_index = pd.read_csv(index_fn)

    def load_task_cfg(problem_id: str) -> DictConfig:
        config_fn = problem_index["config_fn"][problem_index["problem_id"] == problem_id].iloc[0]
        if not Path(config_fn).is_file():
            raise ValueError("Maybe the index is old. Run `python -m carps.utils.index_configs` to refresh.")
        cfg = OmegaConf.load(config_fn)
        return cfg.task

    new_logs = []
    for gid, gdf in logs.groupby(by="problem_id"):
        task_cfg = load_task_cfg(problem_id=gid)
        task_columns = [c for c in gdf.columns if c.startswith("task.")]
        for c in task_columns:
            key = c.split(".")[1]
            if np.nan in gdf[c].unique():
                print(c, key, task_cfg, gid)
                v = task_cfg.get(key)
                if isinstance(v, list | ListConfig):
                    v = [v] * len(gdf)
                gdf[c] = v
        new_logs.append(gdf)
    return pd.concat(new_logs)


def process_logs(logs: pd.DataFrame) -> pd.DataFrame:
    logger.info("Processing raw logs. Normalize n_trials and costs. Calculate trajectory (incumbent cost).")
    # logs= logs.drop(columns=["config"])
    # Filter MO costs
    logs = logs[~logs["problem_id"].str.startswith("DUMMY")]
    logs = logs[~logs["benchmark_id"].str.startswith("DUMMY")]
    logs = logs[~logs["optimizer_id"].str.startswith("DUMMY")]
    logs["trial_value__cost"] = logs["trial_value__cost"].apply(lambda x: x if isinstance(x, float) else eval(x))
    logs = logs[logs["trial_value__cost"].apply(lambda x: isinstance(x, float))]
    logs["trial_value__cost"] = logs["trial_value__cost"].apply(float)
    logs["n_trials_norm"] = logs.groupby("problem_id")["n_trials"].transform(normalize)
    logs["trial_value__cost_norm"] = logs.groupby("problem_id")["trial_value__cost"].transform(normalize)
    logs["trial_value__cost_inc"] = logs.groupby(by=["problem_id", "optimizer_id", "seed"])[
        "trial_value__cost"
    ].transform("cummin")
    logs["trial_value__cost_inc_norm"] = logs.groupby(by=["problem_id", "optimizer_id", "seed"])[
        "trial_value__cost_norm"
    ].transform("cummin")
    logs = maybe_postadd_task(logs)
    if "task.n_objectives" in logs:
        logs["task.is_multiobjective"] = logs["task.n_objectives"] > 1
    logs = add_scenario_type(logs)

    # Add time
    logs = logs.groupby(by=["problem_id", "optimizer_id", "seed"]).apply(calc_time).reset_index(drop=True)
    logs["time_norm"] = logs.groupby("problem_id")["time"].transform(normalize)
    return logs


def calc_time(D: pd.DataFrame) -> pd.Series:
    trialtime = D["trial_value__virtual_time"]
    nulltime = D["trial_value__starttime"] - D["trial_value__starttime"].min()
    trialtime_cum = trialtime.cumsum()
    elapsed = nulltime + trialtime_cum
    elapsed.name = "time"
    D["time"] = elapsed
    return D


def normalize(S: pd.Series, epsilon: float = 1e-8) -> pd.Series:
    return (S - S.min()) / (S.max() - S.min() + epsilon)


def get_interpolated_performance_df(
    logs: pd.DataFrame, n_points: int = 20, x_column: str = "n_trials_norm"
) -> pd.DataFrame:
    """Get performance dataframe for plotting.

    Interpolated at regular intervals.

    Parameters
    ----------
    logs : pd.DataFrame
        Preprocessed logs.
    n_points : int, optional
        Number of interpolation steps, by default 20
    x_column : str, optional
        The x-axis column to interpolate by, by default 'n_trials_norm'

    Raises:
    ------
    ValueError
        When x_column missing in dataframe.

    Returns:
    -------
    pd.DataFrame
        Performance data frame for plotting
    """
    logger.info("Create dataframe for neat plotting by aligning x-axis / interpolating budget.")

    if x_column not in logs:
        msg = f"x_column `{x_column}` not in logs! Did you call `carps.analysis.process_data.process_logs` on the raw logs?"
        raise ValueError(msg)

    interpolation_columns = [
        "trial_value__cost",
        "trial_value__cost_norm",
        "trial_value__cost_inc",
        "trial_value__cost_inc_norm",
    ]
    # interpolation_columns = [
    #     c for c in logs.columns if c != x_column and c not in identifier_columns and not c.startswith("problem")]
    group_keys = ["scenario", "benchmark_id", "optimizer_id", "problem_id", "seed"]
    x = np.linspace(0, 1, n_points + 1)
    D = []
    for gid, gdf in logs.groupby(by=group_keys):
        metadata = dict(zip(group_keys, gid, strict=False))
        performance_data = {}
        performance_data[x_column] = x
        for icol in interpolation_columns:
            if icol in gdf:
                xp = gdf[x_column].to_numpy()
                fp = gdf[icol].to_numpy()
                y = np.interp(x=x, xp=xp, fp=fp)
                performance_data[icol] = y
        performance_data.update(metadata)
        D.append(pd.DataFrame(performance_data))
    return pd.concat(D).reset_index(drop=True)


def load_logs(rundir: str):
    logs_fn = Path(rundir) / "logs.csv"
    logs_cfg_fn = logs_fn.parent / "logs_cfg.csv"

    logger.info(f"Load logs from `{logs_fn}` and associated configs from {logs_cfg_fn}. Preprocess logs.")

    if not logs_fn.is_file() or not logs_cfg_fn.is_file():
        msg = f"No logs found at rundir '{rundir}'. If you used the file logger, you can gather the data with `python -m carps.analysis.gather_data <rundir>`."
        raise RuntimeError(msg)

    df = pd.read_csv(logs_fn)
    df = process_logs(df)
    df_cfg = pd.read_csv(logs_cfg_fn)
    return df, df_cfg


if __name__ == "__main__":
    fire.Fire(load_logs)



"""Pymoo Problem."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING

import numpy as np
import pymoo
import pymoo.problems
from ConfigSpace import ConfigurationSpace, Float
from pymoo.problems.multi.omnitest import OmniTest
from pymoo.problems.multi.sympart import SYMPART, SYMPARTRotated

from carps.benchmarks.problem import Problem
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from carps.loggers.abstract_logger import AbstractLogger

extra_probs = {
    "sympart": SYMPART,
    "sympart_rotated": SYMPARTRotated,
    "omnitest": OmniTest,
}


class PymooProblem(Problem):
    """Pymoo Problem class."""

    def __init__(
        self,
        problem_name: str,
        seed: int,
        metric: str | list[str],
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        """Initialize a Pymoo problem."""
        super().__init__(loggers)

        self.problem_name = problem_name
        self.metric = metric
        if problem_name in extra_probs:
            self._problem = extra_probs[problem_name]()
        else:
            self._problem = pymoo.problems.get_problem(self.problem_name)
        self._configspace = self.get_pymoo_space(pymoo_prob=self._problem, seed=seed)

    @property
    def configspace(self) -> ConfigurationSpace:
        """Return configuration space.

        Returns:
        -------
        ConfigurationSpace
            Configuration space.
        """
        return self._configspace

    def get_pymoo_space(self, pymoo_prob: pymoo.core.problem.Problem, seed: int) -> ConfigurationSpace:
        """Get ConfigSpace from pymoo problem."""
        n_var = pymoo_prob.n_var
        xl, xu = pymoo_prob.xl, pymoo_prob.xu
        hps = [Float(name=f"x{i}", bounds=[xl[i], xu[i]]) for i in range(n_var)]
        configspace = ConfigurationSpace(seed=seed)
        configspace.add_hyperparameters(hps)
        return configspace

    def _evaluate(self, trial_info: TrialInfo) -> TrialValue:
        """Evaluate problem.

        Parameters
        ----------
        trial_info : TrialInfo
            Dataclass with configuration, seed, budget, instance.

        Returns:
        -------
        TrialValue
            Cost
        """
        configuration = np.array(list(trial_info.config.values()))
        start_time = time.time()
        costs = self._problem.evaluate(configuration).tolist()
        if len(costs) == 1:
            costs = costs[0]
        end_time = time.time()

        return TrialValue(
            cost=costs,
            time=end_time - start_time,
            starttime=start_time,
            endtime=end_time,
        )



from __future__ import annotations

from pathlib import Path

import fire
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from carps.analysis.process_data import load_logs
from carps.analysis.utils import filter_only_final_performance
from carps.utils.loggingutils import get_logger

logger = get_logger(__file__)


def custom_latex_table(result, *, decimal_places=3, label=None, only_tabular: bool = True) -> str:
    """Creates a latex table from the results dataframe of the statistical analysis.

    # Parameters

    result (RankResult):
        Should be the return value the autorank function.

    decimal_places (int, default=3):
        Number of decimal places that are used for the report.

    label (str, default=None):
        Label of the table. Defaults to 'tbl:stat_results' if None.
    """
    if label is None:
        label = "tbl:stat_results"

    table_df = result.rankdf
    columns = table_df.columns.to_list()
    if (
        result.omnibus != "bayes"
        and result.pvalue >= result.alpha
        or result.omnibus == "bayes"
        and len({"smaller", "larger"}.intersection(set(result.rankdf["decision"]))) == 0
    ):
        columns.remove("effect_size")
        columns.remove("magnitude")
    if result.posthoc == "tukeyhsd":
        columns.remove("meanrank")
    columns.insert(columns.index("ci_lower"), "CI")
    columns.remove("ci_lower")
    columns.remove("ci_upper")
    rename_map = {}
    if result.effect_size == "cohen_d":
        rename_map["effect_size"] = "$d$"
    elif result.effect_size == "cliff_delta":
        rename_map["effect_size"] = r"D-E-L-T-A"
    elif result.effect_size == "akinshin_gamma":
        rename_map["effect_size"] = r"G-A-M-M-A"
    rename_map["magnitude"] = "Magnitude"
    rename_map["mad"] = "MAD"
    rename_map["median"] = "MED"
    rename_map["meanrank"] = "MR"
    rename_map["mean"] = "M"
    rename_map["std"] = "SD"
    rename_map["decision"] = "Decision"
    format_string = "[{0[ci_lower]:." + str(decimal_places) + "f}, {0[ci_upper]:." + str(decimal_places) + "f}]"
    table_df["CI"] = table_df.agg(format_string.format, axis=1)
    table_df = table_df[columns]
    if result.omnibus == "bayes":
        table_df.at[table_df.index[0], "decision"] = "-"
    table_df = table_df.rename(rename_map, axis="columns")

    float_format = lambda x: ("{:0." + str(decimal_places) + "f}").format(x) if not np.isnan(x) else "-"
    table_string = table_df.to_latex(float_format=float_format, na_rep="-").strip()
    table_string = table_string.replace("D-E-L-T-A", r"$\delta$")
    table_string = table_string.replace("G-A-M-M-A", r"$\gamma$")
    table_string = table_string.replace(r"p\_equal", r"$P(\textit{equal})$")
    table_string = table_string.replace(r"p\_smaller", r"$P(\textit{smaller})$")
    table_string = table_string.replace("_", r"\_")
    final_str = rf"""
\\begin{{table}}[h]
    \\caption{{Summary of populations}}
    \\label{{{label}}}
    \centering
    {table_string}
\end{{table}}
"""

    if only_tabular:
        return table_string
    else:
        return final_str


def get_df_crit(
    df: pd.DataFrame,
    budget_var: str = "n_trials_norm",
    max_budget: float = 1,
    soft: bool = True,
    perf_col: str = "trial_value__cost_inc",
    nan_handling: str = "remove",
) -> pd.DataFrame:
    df = filter_only_final_performance(df=df, budget_var=budget_var, max_budget=max_budget, soft=soft)

    # Work on mean of different seeds
    df_crit = df.groupby(["optimizer_id", "problem_id"])[perf_col].apply(np.nanmean).reset_index()

    df_crit = df_crit.pivot(index="problem_id", columns="optimizer_id", values=perf_col)

    if nan_handling == "remove":
        nan_ids = np.array([np.any(np.isnan(d.values)) for _, d in df_crit.iterrows()])
        lost = df_crit[nan_ids]

        # Rows are problems, cols are optimizers
        df_crit = df_crit[~nan_ids]
        logger.info(f"Lost following experiments: {lost}")
    elif nan_handling == "keep":
        pass
    elif nan_handling == "replace_by_highest":
        nan_ids = np.array([np.any(np.isnan(d.values)) for _, d in df_crit.iterrows()])
        max_val = df_crit[perf_col].max()
        df_crit.loc[nan_ids, perf_col] = max_val
        logger.info(f"Replaced nans by max val {max_val}: {df_crit}")
    else:
        raise ValueError(f"Unknown nan handling {nan_handling}. Can only do `remove`, `keep`, `replace_by_highest`.")

    return df_crit


def calc_critical_difference(
    df: pd.DataFrame, identifier: str | None = None, figsize=(12, 8), perf_col: str = "trial_value__cost_inc_norm"
) -> RankResult:
    df_crit = get_df_crit(df, perf_col=perf_col)

    # result = autorank(df_crit, alpha=0.05, verbose=True)
    # create_report(result)

    # fig, ax = plt.subplots(figsize=(6,2))
    # ax = plot_stats(result, ax=ax)

    # if identifier is None:
    #     identifier = ""
    # else:
    #     identifier = "_" + identifier
    # fn = f"figures/critd/criticaldifference{identifier}"
    # savefig(fig=fig, filename=fn + ".png")
    # savefig(fig=fig, filename=fn + ".pdf")

    return cd_evaluation(
        df_crit,
        maximize_metric=False,
        ignore_non_significance=True,
        output_path=f"figures/critd/cd{identifier}",
        figsize=figsize,
    )


def calc(rundir: str, scenario: str = "blackbox") -> None:
    df, df_cfg = load_logs(rundir=rundir)
    calc_critical_difference(df=df[df["scenario"] == scenario], identifier=scenario)


"""Code for CD Plots
Requirements: the usual (pandas, matplotlib, numpy) and autorank==1.1.3

Source: https://gist.github.com/LennartPurucker/cf4616512529e29c123608b6c2c4a7e9
"""

import math
import warnings

from autorank._util import RankResult, get_sorted_rank_groups, rank_multiple_nonparametric, test_normality


def _custom_cd_diagram(result, reverse, ax, width):
    """!TAKEN FROM AUTORANK WITH MODIFICATIONS!"""

    def plot_line(line, color="k", **kwargs):
        ax.plot([pos[0] / width for pos in line], [pos[1] / height for pos in line], color=color, **kwargs)

    def plot_text(x, y, s, *args, **kwargs):
        ax.text(x / width, y / height, s, *args, **kwargs)

    sorted_ranks, names, groups = get_sorted_rank_groups(result, reverse)
    cd = result.cd

    lowv = min(1, int(math.floor(min(sorted_ranks))))
    highv = max(len(sorted_ranks), int(math.ceil(max(sorted_ranks))))
    cline = 0.4
    textspace = 1
    scalewidth = width - 2 * textspace

    def rankpos(rank):
        relative_rank = rank - lowv if not reverse else highv - rank
        return textspace + scalewidth / (highv - lowv) * relative_rank

    linesblank = 0.2 + 0.2 + (len(groups) - 1) * 0.1

    # add scale
    distanceh = 0.1
    cline += distanceh

    # calculate height needed height of an image
    minnotsignificant = max(2 * 0.2, linesblank)
    height = cline + ((len(sorted_ranks) + 1) / 2) * 0.2 + minnotsignificant

    if ax is None:
        fig = plt.figure(figsize=(width, height))
        fig.set_facecolor("white")
        ax = fig.add_axes([0, 0, 1, 1])  # reverse y axis
    ax.set_axis_off()

    # Upper left corner is (0,0).3
    ax.plot([0, 1], [0, 1], c="w")
    ax.set_xlim(0, 1)
    ax.set_ylim(1, 0)

    plot_line([(textspace, cline), (width - textspace, cline)], linewidth=0.7)

    bigtick = 0.1
    smalltick = 0.05

    tick = None
    for a in [*list(np.arange(lowv, highv, 0.5)), highv]:
        tick = smalltick
        if a == int(a):
            tick = bigtick
        plot_line([(rankpos(a), cline - tick / 2), (rankpos(a), cline)], linewidth=0.7)

    for a in range(lowv, highv + 1):
        plot_text(rankpos(a), cline - tick / 2 - 0.05, str(a), ha="center", va="bottom")

    for i in range(math.ceil(len(sorted_ranks) / 2)):
        chei = cline + minnotsignificant + i * 0.2
        plot_line(
            [(rankpos(sorted_ranks[i]), cline), (rankpos(sorted_ranks[i]), chei), (textspace - 0.1, chei)],
            linewidth=0.7,
        )
        plot_text(textspace - 0.2, chei, names[i], ha="right", va="center")

    for i in range(math.ceil(len(sorted_ranks) / 2), len(sorted_ranks)):
        chei = cline + minnotsignificant + (len(sorted_ranks) - i - 1) * 0.2
        plot_line(
            [(rankpos(sorted_ranks[i]), cline), (rankpos(sorted_ranks[i]), chei), (textspace + scalewidth + 0.1, chei)],
            linewidth=0.7,
        )
        plot_text(textspace + scalewidth + 0.2, chei, names[i], ha="left", va="center")

    # upper scale
    if not reverse:
        begin, end = rankpos(lowv), rankpos(lowv + cd)
    else:
        begin, end = rankpos(highv), rankpos(highv - cd)
    distanceh += 0.15
    bigtick /= 2
    plot_line([(begin, distanceh), (end, distanceh)], linewidth=0.7)
    plot_line([(begin, distanceh + bigtick / 2), (begin, distanceh - bigtick / 2)], linewidth=0.7)
    plot_line([(end, distanceh + bigtick / 2), (end, distanceh - bigtick / 2)], linewidth=0.7)
    plot_text((begin + end) / 2, distanceh - 0.05, "CD", ha="center", va="bottom")

    # no-significance lines
    side = 0.05
    no_sig_height = 0.1
    start = cline + 0.2
    for l, r in groups:
        plot_line([(rankpos(sorted_ranks[l]) - side, start), (rankpos(sorted_ranks[r]) + side, start)], linewidth=2.5)
        start += no_sig_height

    return ax


def cd_evaluation(
    performance_per_dataset,
    maximize_metric,
    output_path=None,
    ignore_non_significance=False,
    plt_title=None,
    figsize=(12, 8),
) -> RankResult:
    """Performance per dataset is  a dataframe that stores the performance (with respect to a metric) for  set of
    configurations / models / algorithms per dataset. In  detail, the columns are individual configurations.
    rows are datasets and a cell is the performance of the configuration for  dataset.
    """
    # -- Preprocess data for autorank
    rank_data = performance_per_dataset.copy() * -1 if maximize_metric else performance_per_dataset.copy()
    rank_data = rank_data.reset_index(drop=True)
    rank_data = pd.DataFrame(rank_data.values, columns=list(rank_data))

    # -- Settings for autorank
    alpha = 0.05
    effect_size = None
    verbose = True
    order = "ascending"  # always due to the preprocessing
    alpha_normality = alpha / len(rank_data.columns)
    all_normal, pvals_shapiro = test_normality(rank_data, alpha_normality, verbose)

    # -- Friedman-Nemenyi
    res = rank_multiple_nonparametric(rank_data, alpha, verbose, all_normal, order, effect_size, None)

    result = RankResult(
        res.rankdf,
        res.pvalue,
        res.cd,
        res.omnibus,
        res.posthoc,
        all_normal,
        pvals_shapiro,
        None,
        None,
        None,
        alpha,
        alpha_normality,
        len(rank_data),
        None,
        None,
        None,
        None,
        res.effect_size,
        None,
    )
    print(res.rankdf)
    is_significant = True
    if result.pvalue >= result.alpha:
        if ignore_non_significance:
            warnings.warn("Result is not significant and results of the plot may be misleading!")
            is_significant = False
        else:
            raise ValueError(
                "Result is not significant and results of the plot may be misleading. If you still want to see the CD plot, set"
                + " ignore_non_significance to True."
            )

    # -- Plot
    fig, ax = plt.subplots(figsize=figsize)
    plt.rcParams.update({"font.size": 16})
    _custom_cd_diagram(result, False, ax, figsize[0])  # order == "descending", ax, 8)
    if plt_title or not is_significant:
        plt_title = ""
        if not is_significant:
            plt_title += "(non-significant)"
        plt.title(plt_title)
    plt.tight_layout()
    if output_path:
        Path(output_path).parent.mkdir(exist_ok=True, parents=True)
        plt.savefig(output_path + ".png", transparent=True, bbox_inches="tight")
        plt.savefig(output_path + ".pdf", transparent=True, bbox_inches="tight")

    plt.show()
    plt.close()

    return result


# if __name__ == "__main__":
#     input_data = pd.read_csv("./benchmark_binary_1hour.csv")  # assume table format where each row represents some experiment run

#     # Pivot for desired metric to create the performance per dataset table
#     performance_per_dataset = input_data.pivot(index="task", columns="framework", values="result")

#     # shows the plot by default
#     cd_evaluation(performance_per_dataset, maximize_metric=True, output_path=None, ignore_non_significance=False)

if __name__ == "__main__":
    fire.Fire(calc)



from __future__ import annotations
import pandas as pd
from carps.analysis.gather_data import load_set, convert_mixed_types_to_str

def concat_rundata():
    paths = {
        "BBsubset": {
            "dev": ["runs_subset_BB/dev"],
            "test": ["runs_subset_BB/test"],
        },
        "MFsubset": {
            "dev": ["runs_subset_MF/dev"],
            "test": ["runs_subset_MF/test"],
        },
        "MOsubset": {
            "dev": ["runs_subset_MO/dev"],
            "test": ["runs_subset_MO/test"],
        },
        "MOMFsubset": {
            "dev": ["runs_subset_MOMF/dev"],
            "test": ["runs_subset_MOMF/test"],
        },
    }

    args = []
    for item in paths.values():
        for k,v in item.items():
            from pathlib import Path
            args.append((v,k))
    res = [load_set(paths=a[0], set_id=a[1]) for a in args]
    df = pd.concat([r[0] for r in res]).reset_index(drop=True)
    df = convert_mixed_types_to_str(df)
    df.to_parquet("rundata.parquet")

if __name__ == "__main__":
    concat_rundata()


from __future__ import annotations

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


def norm_by_opt(df: pd.DataFrame, optimizer_id: str) -> pd.DataFrame:
    df_new = []
    for _gid, gdf in df.groupby(by=["problem_id", "seed"]):
        reference = gdf[gdf["optimizer_id"] == optimizer_id]["trial_value__cost"]
        gdf["trial_value__cost_normopt"] = gdf.groupby("optimizer_id")["trial_value__cost"].transform(
            lambda x: x / reference
        )
        df_new.append(gdf)
    df = pd.concat(df_new).reset_index(drop=True)

    # df["trial_value_cost_normopt"] = df.groupby("problem_id").apply(_norm_by_opt)
    df["trial_value__cost_inc_normopt"] = df.groupby(by=["problem_id", "optimizer_id", "seed"])[
        "trial_value__cost_normopt"
    ].transform("cummin")
    return df


if __name__ == "__main__":
    df = pd.read_parquet(
        "/scratch/hpc-prf-intexml/cbenjamins/repos/CARP-S-Experiments/lib/CARP-S/runs/RandomSearch/MFPBench/logs.parquet"
    )

    df_new = df.copy()
    df_new["optimizer_id"] = "HeheOpt"
    df_new["trial_value__cost"] -= 0.1
    df = pd.concat([df, df_new]).reset_index(drop=True)

    fig, ax = plt.subplots(figsize=(6, 4))
    normalize_by_opt = "RandomSearch"
    df = norm_by_opt(df, "RandomSearch")

    sns.lineplot()



"""Yahpo problem class."""

from __future__ import annotations

import time
from pathlib import Path
from typing import TYPE_CHECKING

from omegaconf import ListConfig
from yahpo_gym import BenchmarkSet, list_scenarios, local_config

from carps.benchmarks.problem import Problem
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import ConfigurationSpace

    from carps.loggers.abstract_logger import AbstractLogger

LOWER_IS_BETTER = {
    "mmce": True,  # classification error
    "f1": False,
    "auc": False,
    "logloss": True,
    "nf": True,  # number of features used
    "ias": True,  # interaction strength of features  # TODO check
    "rammodel": True,  # model size
    "val_accuracy": False,
    "val_cross_entropy": True,
    "acc": False,
    "bac": False,  # balanced acc
    "brier": True,
    "memory": True,
    "timetrain": True,
    "runtime": True,
    "mec": True,  # main effect complexity of features
    "valid_mse": True,
}


def maybe_invert(value: float, target: str) -> float:
    sign = 1
    if not LOWER_IS_BETTER[target]:
        sign = -1
    return sign * value


class YahpoProblem(Problem):
    """Yahpo Problem."""

    def __init__(
        self,
        bench: str,
        instance: str,
        metric: str | list[str],
        budget_type: str | None = None,
        lower_is_better: bool = True,
        yahpo_data_path: str | None = None,
        loggers: list[AbstractLogger] | None = None,
    ):
        """Initialize a Yahpo problem.

        Parameters
        ----------
        bench: str
            Benchmark name.
        instance : str
            Instance name.
        metric : str
            Metric(s) to optimize for (depends on the Benchmark instance e.g. lcbench).
        budget_type : Optional[str]
            Budget type for the multifidelity setting. Should be None for the blackbox setting.
        lower_is_better: bool
            Whether the metric is to be minimized or maximized.
        yahpo_data_path : str | None
            Path to yahpo data, defaults to '../benchmark_data/yahpo_data' (relative to this file).
        """
        super().__init__(loggers)

        assert bench in list_scenarios(), f"The scenario {bench} you choose is not available."

        yahpo_data_path = yahpo_data_path or Path(__file__).parent.parent / "benchmark_data/yahpo_data"

        # setting up meta data for surrogate benchmarks
        local_config.init_config()
        local_config.set_data_path(yahpo_data_path)

        self.scenario = bench
        self.instance = str(instance)

        self._problem = BenchmarkSet(scenario=bench, instance=self.instance, check=False)
        self._configspace = self._problem.get_opt_space(drop_fidelity_params=True)
        self.fidelity_space = self._problem.get_fidelity_space()
        self.fidelity_dims = list(self._problem.get_fidelity_space()._hyperparameters.keys())

        self.budget_type = budget_type
        self.lower_is_better = lower_is_better

        assert self.budget_type in [*self.fidelity_dims, None], (
            f"The budget type {self.budget_type} you choose is "
            f"not available in this instance. Please choose "
            f"from {[*self.fidelity_dims, None]}."
        )

        if self.budget_type is None or len(self.fidelity_dims) > 1:
            other_fidelities = [fid for fid in self.fidelity_dims if fid != self.budget_type]
            self.max_other_fidelities = {}
            for fidelity in other_fidelities:
                self.max_other_fidelities[fidelity] = self.fidelity_space.get_hyperparameter(fidelity).upper

        if not isinstance(metric, list | ListConfig):
            metric = [metric]
        self.metrics = metric

    @property
    def configspace(self) -> ConfigurationSpace:
        """Return configuration space.

        Returns:
        -------
        ConfigurationSpace
            Configuration space.
        """
        return self._configspace

    # @property
    # FIXME: see caro's message:
    #  the idea is somehow to overwrite the optimizer/multifidelity attributes for
    #  budget_variable and min_budget, max_budget with a FidelitiySpace class without interpolation
    #  that is based on the problem instance / config file. Similarly find out how to deal with
    #  the metrics.
    # def fidelity_space(self):
    #     return FidelitySpace(self.fidelity_dims)

    def _evaluate(self, trial_info: TrialInfo) -> TrialValue:
        """Evaluate problem.

        Parameters
        ----------
        trial_info : TrialInfo
            Dataclass with configuration, seed, budget, instance.

        Returns:
        -------
        TrialValue
            Cost
        """
        configuration = trial_info.config
        xs = configuration.get_dictionary()

        # If there are multiple fidelities, we take maximum fidelity value for the respective other fidelity dimensions
        # If we are in the blackbox setting, we take maximum fidelity value for all fidelity dimensions
        starttime = time.time()
        if self.budget_type is not None:
            if self.budget_type == "trainsize":
                xs.update({self.budget_type: trial_info.budget})
            elif trial_info.budget is not None:  # to avoid mypy error
                xs.update({self.budget_type: round(trial_info.budget)})

        if self.budget_type is None or len(self.fidelity_dims) > 1:
            xs.update(self.max_other_fidelities)

        # Benchmarking suite returns a list of results (as potentially more than one config can be passed),
        # as we only pass one config we need to select the first one
        ret = self._problem.objective_function(configuration=xs, seed=trial_info.seed)[0]
        costs = [maybe_invert(ret[target], target) for target in self.metrics]
        virtual_time = ret.get("time", 0.0)
        if len(costs) == 1:
            costs = costs[0]

        endtime = time.time()
        T = endtime - starttime

        return TrialValue(cost=costs, time=T, starttime=starttime, endtime=endtime, virtual_time=virtual_time)



from __future__ import annotations

from time import sleep
from typing import TYPE_CHECKING, Any

from carps.optimizers.optimizer import Optimizer
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import Configuration, ConfigurationSpace
    from omegaconf import DictConfig

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent, SearchSpace


class DummyOptimizer(Optimizer):
    def __init__(
        self,
        dummy_cfg: DictConfig,
        problem: Problem,
        task: Task,
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        super().__init__(problem, task, loggers)
        self.cfg = dummy_cfg
        self.history: list[TrialInfo, TrialValue] = []
        if "budget" in self.cfg:
            self.fidelity_enabled = True
            self.budget = self.cfg.budget
        else:
            self.fidelity_enabled = False
            self.budget = None

    def _setup_optimizer(self) -> Any:
        pass

    def convert_configspace(self, configspace: ConfigurationSpace) -> SearchSpace:
        return configspace

    def convert_to_trial(self, config: Configuration, budget: float | None = None) -> TrialInfo:
        return TrialInfo(config=config, budget=budget)

    def ask(self) -> TrialInfo:
        sleep(1)
        config = self.problem.configspace.sample_configuration()
        return self.convert_to_trial(config, budget=self.budget)

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        self.history.append((trial_info, trial_value))

    def get_current_incumbent(self) -> Incumbent:
        return min(self.history, key=lambda x: x[1].cost)



from __future__ import annotations

from typing import TYPE_CHECKING

import hydra
from hydra.core.hydra_config import HydraConfig

from carps.utils.loggingutils import get_logger, setup_logging
from carps.utils.requirements import check_requirements
from carps.utils.running import optimize

if TYPE_CHECKING:
    from omegaconf import DictConfig

setup_logging()
logger = get_logger(__file__)


@hydra.main(config_path="configs", config_name="base.yaml", version_base=None)  # type: ignore[misc]
def main(cfg: DictConfig) -> None:
    """Run optimizer on problem.

    Save trajectory and metadata to database.

    Parameters
    ----------
    cfg : DictConfig
        Global configuration.

    """
    hydra_cfg = HydraConfig.instance().get()
    overrides = hydra_cfg.overrides.task
    logger.info(f"Runcommand: `python -m carps.run {' '.join(overrides)}`")
    check_requirements(cfg=cfg)
    optimize(cfg=cfg)


if __name__ == "__main__":
    main()



from __future__ import annotations

import copy
import datetime
from collections import OrderedDict
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import numpy as np
import omegaconf
from ConfigSpace import Configuration, ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    FloatHyperparameter,
    Hyperparameter,
    IntegerHyperparameter,
    OrdinalHyperparameter,
)
from syne_tune.backend.trial_status import (
    Status,
    Trial as SyneTrial,
    TrialResult,
)
from syne_tune.config_space import (
    choice,
    lograndint,
    loguniform,
    ordinal,
    randint,
    uniform,
)
from syne_tune.optimizer.baselines import (
    ASHA,
    BOHB,
    BORE,
    DEHB,
    KDE,
    MOASHA,
    MOBSTER,
    SyncMOBSTER,
    MOREA,
    BayesianOptimization,
    MOLinearScalarizationBayesOpt,
    MORandomScalarizationBayesOpt,
)

from carps.optimizers.optimizer import Optimizer
from carps.utils.pareto_front import pareto
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from syne_tune.optimizer.scheduler import TrialScheduler as SyneTrialScheduler

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent

# This is a subset from the syne-tune baselines
optimizers_dict = {
    "BayesianOptimization": BayesianOptimization,
    "BO-MO-RS": MORandomScalarizationBayesOpt,
    "BO-MO-LS": MOLinearScalarizationBayesOpt,
    "MOREA": MOREA,
    "ASHA": ASHA,
    "MOBSTER": MOBSTER,
    "BOHB": BOHB,
    "KDE": KDE,
    "BORE": BORE,
    "DEHB": DEHB,
    "MOASHA": MOASHA,
    "SyncMOBSTER": SyncMOBSTER,
}

mf_optimizer_dicts = {
    "with_mf": {"ASHA", "MOASHA", "DEHB", "MOBSTER", "BOHB", "SyncMOBSTER"},
    "without_mf": {"BORE", "BayesianOptimization", "KDE"},
}


def configspaceHP2syneTuneHP(hp: Hyperparameter) -> Callable:
    if isinstance(hp, IntegerHyperparameter):
        if hp.log:
            return lograndint(hp.lower, hp.upper)
        else:
            return randint(hp.lower, hp.upper)
    elif isinstance(hp, FloatHyperparameter):
        if hp.log:
            return loguniform(hp.lower, hp.upper)
        else:
            return uniform(hp.lower, hp.upper)
    elif isinstance(hp, CategoricalHyperparameter):
        return choice(hp.choices)
    elif isinstance(hp, OrdinalHyperparameter):
        return ordinal(hp.sequence)
    elif isinstance(hp, Constant):
        return choice([hp.value])
    else:
        raise NotImplementedError(f"Unknown hyperparameter type: {hp.__class__.__name__}")


class SynetuneOptimizer(Optimizer):
    def __init__(
        self,
        problem: Problem,
        optimizer_name: str,
        task: Task,
        optimizer_kwargs: dict | None = None,
        loggers: list[AbstractLogger] | None = None,
        conversion_factor: int = 1000,
    ) -> None:
        super().__init__(problem, task, loggers)
        self.fidelity_enabled = False
        self.max_budget = task.max_budget
        assert optimizer_name in optimizers_dict
        if optimizer_name in mf_optimizer_dicts["with_mf"]:
            # raise NotImplementedError("Multi-Fidelity Optimization on SyneTune is not implemented yet!")
            self.fidelity_enabled = True
            if self.task.fidelity_type is None:
                raise ValueError("To run multi-fidelity optimizer, the problem must define a fidelity type!")
            if self.max_budget is None:
                raise ValueError("To run multi-fidelity optimizer, we must specify max_budget!")

        self.fidelity_type: str = self.task.fidelity_type
        self.configspace = self.problem.configspace
        self.metric: str | list[str] = self.task.objectives
        if len(self.metric) == 1:
            self.metric = self.metric[0]
        self.conversion_factor = conversion_factor
        self.trial_counter: int = 0

        self.optimizer_name = optimizer_name
        self._solver: SyneTrialScheduler | None = None

        self.optimizer_kwargs = (
            omegaconf.OmegaConf.to_object(optimizer_kwargs) if optimizer_kwargs is not None else None
        )

        self.completed_experiments: OrderedDict[int, TrialResult] = OrderedDict()
        self.convert = False

    def convert_configspace(self, configspace: ConfigurationSpace) -> dict[str, Any]:
        """Convert configuration space from Problem to Optimizer.

        Convert the configspace from ConfigSpace to syne-tune. However, given that syne-tune does not support
        conditions and forbidden clauses, we only add hyperparameters here

        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space from Problem.

        dict[str, Any]
        -------
        configspace_st
            Configuration space for syne tune.
        """
        configspace_st = {}
        for k, v in configspace.items():
            configspace_st[k] = configspaceHP2syneTuneHP(v)
        if self.fidelity_enabled:
            configspace_st[self.fidelity_type] = (
                self.max_budget if not self.convert else int(self.conversion_factor * self.max_budget)
            )

        return configspace_st

    def convert_to_trial(self, trial: SyneTrial) -> TrialInfo:  # type: ignore[override]
        """Convert proposal from SyneTune to TrialInfo.

        Parameters
        ----------
        config : Configuration
            Configuration
        seed : int | None, optional
            Seed, by default None
        budget : float | None, optional
            Budget, by default None
        instance : str | None, optional
            Instance, by default None

        Returns:
        -------
        TrialInfo
            Trial info containing configuration, budget, seed, instance.
        """
        configs = copy.deepcopy(trial.config)
        budget = configs.pop(self.fidelity_type) if self.fidelity_type is not None else None
        configuration = Configuration(
            configuration_space=self.configspace, values=configs, allow_inactive_with_values=True
        )
        return TrialInfo(
            config=configuration,
            seed=None,
            budget=budget if not self.convert else budget / self.conversion_factor,
            instance=None,
        )

    def ask(self) -> TrialInfo:
        """Ask the optimizer for a new trial to evaluate.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Returns:
        -------
        TrialInfo
            trial info (config, seed, instance, budget)
        """
        trial_suggestion = self._solver.suggest(self.trial_counter)
        trial = SyneTrial(
            trial_id=self.trial_counter,
            config=trial_suggestion.config,
            creation_time=datetime.datetime.now(),
        )
        return self.convert_to_trial(trial=trial)

    def convert_to_synetrial(self, trial_info: TrialInfo) -> SyneTrial:
        """Convert a trial info to the syne tune format.

        Parameters
        ----------
        trial_info : TrialInfo
            Trial info.

        Returns:
        -------
        SyneTrial
            Synetune format
        """
        syne_config = dict(trial_info.config)
        if self.fidelity_enabled:
            syne_config[self.fidelity_type] = (
                trial_info.budget if not self.convert else int(self.conversion_factor * trial_info.budget)
            )
        return SyneTrial(
            trial_id=self.trial_counter,
            config=syne_config,
            creation_time=datetime.datetime.now(),
        )

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue):
        """Tell the optimizer a new trial.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Parameters
        ----------
        trial_info : TrialInfo
            trial info (config, seed, instance, budget)
        trial_value : TrialValue
            trial value (cost, time, ...)
        """
        cost = trial_value.cost
        trial = self.convert_to_synetrial(trial_info=trial_info)
        experiment_result = {}
        if self.task.n_objectives == 1:
            experiment_result = {self.task.objectives[0]: cost}
        else:
            experiment_result = {self.task.objectives[i]: cost[i] for i in range(len(cost))}

        if self.task.is_multifidelity:
            experiment_result[self.fidelity_type] = (
                trial_info.budget if not self.convert else int(self.conversion_factor * trial_info.budget)
            )
            # del experiment_result[self.task.objectives]

            self._solver.on_trial_add(trial=trial)
        self.trial_counter += 1

        self._solver.on_trial_complete(trial=trial, result=experiment_result)
        trial_result = trial.add_results(
            metrics=experiment_result,
            status=Status.completed,
            training_end_time=datetime.datetime.now(),
        )
        self.completed_experiments[trial_result.trial_id] = trial_result

    def best_trial(self, metric: str) -> TrialResult:
        """Return the best trial according to the provided metric."""
        if self.optimizer_name == "MOASHA":
            self.solver.mode = "min"

        sign = 1.0 if self.solver.mode == "max" else -1.0

        return max(
            [value for key, value in self.completed_experiments.items()],
            key=lambda trial: sign * trial.metrics[metric],
        )

    def get_pareto_front(self) -> list[TrialResult]:
        """Return the pareto front for multi-objective optimization."""
        if self.task.is_multifidelity:
            # Determine maximum budget run
            max_budget = np.max([v.metrics[self.fidelity_type] for v in self.completed_experiments.values()])
            # Get only those trial results that ran on max budget
            results_on_highest_fidelity = np.array(
                [v for v in self.completed_experiments.values() if v.metrics[self.fidelity_type] == max_budget]
            )
            # Get costs, exclude fidelity
            costs = np.array([[v.metrics[m] for m in self.task.objectives] for v in results_on_highest_fidelity])
            # Determine pareto front of the trials run on max budget
            front = results_on_highest_fidelity[pareto(costs)]
        else:
            results = np.array(list(self.completed_experiments.values()))
            costs = np.array([list(trial.metrics.values()) for trial in results])
            front = results[pareto(costs)]
        return front.tolist()

    def _setup_optimizer(self) -> SyneTrialScheduler:
        """Setup Optimizer.

        Retrieve defaults and instantiate SyneTune.

        Returns:
        -------
        SyneTrialScheduler
            Instance of a SyneTune.

        """
        if self.optimizer_kwargs is None:
            self.optimizer_kwargs = {}

        _optimizer_kwargs = {
            "metric": self.metric,
            "mode": "min" if self.task.n_objectives == 1 else list(np.repeat("min", self.task.n_objectives)),
        }

        if self.optimizer_name in mf_optimizer_dicts["with_mf"]:
            _optimizer_kwargs["resource_attr"] = self.fidelity_type
            # _optimizer_kwargs["max_t"] = self.max_budget  # TODO check how to set n trials / wallclock limit for synetune

            # for floating point resources like trainsize, we need to convert them to integer for synetune
            if "grace_period" in self.optimizer_kwargs and isinstance(self.optimizer_kwargs["grace_period"], float):
                self.convert = True
                _optimizer_kwargs["grace_period"] = int(self.conversion_factor * self.optimizer_kwargs["grace_period"])

                if "max_t" in self.optimizer_kwargs:
                    _optimizer_kwargs["max_t"] = int(self.conversion_factor * self.max_budget)

                if "max_resource_level" in self.optimizer_kwargs:
                    _optimizer_kwargs["max_resource_level"] = int(
                        self.conversion_factor * self.optimizer_kwargs["max_resource_level"]
                    )

            print(_optimizer_kwargs)

        self.syne_tune_configspace = self.convert_configspace(self.configspace)
        _optimizer_kwargs["config_space"] = self.syne_tune_configspace

        self.optimizer_kwargs.update(_optimizer_kwargs)

        if self.optimizer_name == "MOASHA":
            self.metric = self.optimizer_kwargs["metric"]
            del self.optimizer_kwargs["metric"]
            del self.optimizer_kwargs["resource_attr"]

        if self.optimizer_name in ["SyncMOBSTER"]:
            # del self.optimizer_kwargs["metric"]
            if "time_attr" in self.optimizer_kwargs:
                del self.optimizer_kwargs["time_attr"]

        return optimizers_dict[self.optimizer_name](**self.optimizer_kwargs)

    def get_current_incumbent(self) -> Incumbent:
        if self.task.n_objectives == 1:
            trial_result = self.best_trial(metric=self.task.objectives[0])
            trial_info = self.convert_to_trial(trial=trial_result)
            cost = trial_result.metrics[self.task.objectives[0]]
            trial_value = TrialValue(
                cost=cost,
                time=trial_result.seconds,
                virtual_time=trial_result.seconds,
                additional_info=trial_result.metrics,
            )
            incumbent_tuple = (trial_info, trial_value)
        else:
            trial_result = self.get_pareto_front()
            tis, tvs = [], []
            for result in trial_result:
                trial_info = self.convert_to_trial(trial=result)
                costs = list(result.metrics.values())
                trial_value = TrialValue(
                    cost=costs, time=result.seconds, virtual_time=result.seconds, additional_info=result.metrics
                )
                tis.append(trial_info)
                tvs.append(trial_value)
            incumbent_tuple = list(zip(tis, tvs, strict=False))
        return incumbent_tuple



# ml system singularity
from __future__ import annotations

from carps.utils.running import make_problem
from carps.utils.trials import TrialInfo
from omegaconf import OmegaConf

fns = [
    "carps/configs/problem/HPOBench/blackbox/surr/cfg_surr_ParamNet_Adult.yaml",
    "carps/configs/problem/HPOBench/blackbox/surr/cfg_surr_SVM_default.yaml",
    "carps/configs/problem/HPOBench/blackbox/tab/cfg_ml_rf_7592.yaml",
    "carps/configs/problem/HPOBench/multifidelity/cfg_ml_lr_9981_subsample.yaml",
]

seed = 1
for fn in fns:
    cfg = OmegaConf.load(fn)

    cfg.problem.seed = seed
    if hasattr(cfg.problem, "problem"):
        cfg.problem.problem.rng = seed
    print(cfg)
    problem = make_problem(cfg=cfg)
    print(problem)

    config = problem.configspace.sample_configuration()
    res = [problem.evaluate(TrialInfo(config=config)).cost for _ in range(5)]
    print(res)
    del problem



from __future__ import annotations

from collections import Counter
from pathlib import Path

import pytest
from omegaconf import OmegaConf


@pytest.mark.parametrize(
    ("path", "key"),
    [
        ("carps/configs/optimizer", "optimizer_id"),
        ("carps/configs/problem", "problem_id"),
    ],
)
def test_unique_ids(path, key):
    path = Path(path)
    paths = list(path.glob("**/*.yaml"))
    values = []
    for p in paths:
        cfg = OmegaConf.load(p)
        value = cfg.get(key)
        values.append(value)

    assert len(set(values)) == len(values), (
        "Duplicate " f"{key}, they need to have a unique name: {[k for k, v in Counter(values).items() if v > 1]}"
    )



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

import fire
import numpy as np
import pandas as pd
from pymoo.indicators.hv import HV

from carps.analysis.gather_data import convert_mixed_types_to_str

run_id = ["scenario", "benchmark_id", "problem_id", "optimizer_id", "seed"]


def gather_trajectory(x: pd.DataFrame) -> pd.DataFrame:
    metadata = dict(zip(run_id, x.name, strict=False))
    data = []
    for n_trials, gdf in x.groupby("n_trials"):
        cost_inc = (
            gdf["trial_value__cost_raw"].apply(eval).apply(lambda x: np.array([np.array(c) for c in x])).to_numpy()
        )
        n_obj = len(cost_inc[0])
        cost_inc = np.concatenate(cost_inc).reshape(-1, n_obj)
        D = {
            "n_trials": n_trials,
            "n_incumbents": len(gdf),
            "trial_value__cost": cost_inc,
            "trial_value__cost_inc": cost_inc,
        }
        D.update(metadata)
        data.append(D)
    return pd.DataFrame(data)


def add_reference_point(x: pd.DataFrame) -> pd.DataFrame:
    costs = x["trial_value__cost_inc"].apply(lambda x: np.array([np.array(c) for c in x])).to_list()
    costs = np.concatenate(costs)
    reference_point = np.max(costs, axis=0)
    x["reference_point"] = [reference_point] * len(x)
    return x


def calc_hv(x: pd.DataFrame) -> pd.DataFrame:
    F = np.concatenate(np.array([np.array(p) for p in x["trial_value__cost_inc"].to_numpy()]))

    ind = HV(ref_point=x["reference_point"].iloc[0], pf=None, nds=False)
    x["hypervolume"] = ind(F)
    return x


def serialize_array(arr: np.ndarray):
    return json.dumps(arr.tolist())


def deserialize_array(serialized_arr):
    deserialized = serialized_arr
    try:
        deserialized = np.array(json.loads(serialized_arr))
        print(deserialized)
    except Exception as e:
        print(e)
        print(serialized_arr)
    return deserialized


def maybe_serialize(x: Any) -> Any:
    if isinstance(x, np.ndarray):
        return serialize_array(x)
    else:
        return x


def maybe_deserialize(x: Any) -> Any:
    if isinstance(x, str):
        return deserialize_array(x)
    else:
        return x


def calculate_hypervolume(rundir: str) -> None:
    fn = Path(rundir) / "logs.parquet"
    if not fn.is_file():
        raise ValueError(
            f"Cannot find {fn}. Did you run `python -m carps.analysis.gather_data {rundir} trajectory_logs.jsonl`?"
        )
    df = pd.read_parquet(fn)
    if df["scenario"].nunique() > 2 or df["scenario"].unique()[0] != "multi-objective":
        raise ValueError(f"Oops, found some non multi-objective logs in {fn}. This might not work...")
    trajectory_df = df.groupby(by=run_id).apply(gather_trajectory).reset_index(drop=True)
    trajectory_df = (
        trajectory_df.groupby(by=["scenario", "problem_id"]).apply(add_reference_point).reset_index(drop=True)
    )
    trajectory_df = trajectory_df.groupby(by=[*run_id, "n_trials"]).apply(calc_hv).reset_index(drop=True)
    trajectory_df.to_csv(Path(rundir) / "trajectory.csv")
    trajectory_df = convert_mixed_types_to_str(trajectory_df)
    trajectory_df.to_parquet(Path(rundir) / "trajectory.parquet")


def load_trajectory(rundir: str) -> pd.DataFrame:
    fn = Path(rundir) / "trajectory.parquet"
    if not fn.is_file():
        raise ValueError(f"Cannot find {fn}. Did you run `python -m carps.analysis.calc_hypervolume {rundir}`?")
    df = pd.read_parquet(fn)
    df = df.map(maybe_deserialize)
    print(df["trial_value__cost"].iloc[0], type(df["trial_value__cost"].iloc[0]))


if __name__ == "__main__":
    fire.Fire(calculate_hypervolume)



import datetime

name = "CARP-S"
package_name = "carps"
author = "AutoML"
author_email = "smac@ai.uni-hannover.de"
description = "Benchmark SMAC"
url = "https://www.automl.org"
project_urls = {
    "Documentation": "https://AutoML.github.io/SMACBenchmarking/main",
    "Source Code": "https://github.com/AutoML/carps",
}
copyright = f"Copyright {datetime.date.today().strftime('%Y')}, AutoML"
version = "0.0.1"



"""Run gather data as a slurm job with many cpus.

Example command:
python -m carps.analysis.gather_data_slurm 'rundir=runs/DEHB,runs/SMAC3-Hyperband,runs/SMAC3-MultiFidelityFacade' -m

"""

from __future__ import annotations

from typing import TYPE_CHECKING

import hydra

from carps.analysis.gather_data import filelogs_to_df

if TYPE_CHECKING:
    import pandas as pd
    from omegaconf import DictConfig


@hydra.main(config_path="configs", config_name="gather_data_slurm.yaml", version_base=None)  # type: ignore[misc]
def main(cfg: DictConfig) -> tuple[pd.DataFrame, pd.DataFrame]:
    return filelogs_to_df(cfg.rundir)


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING

from omegaconf import OmegaConf
from py_experimenter.experiment_status import ExperimentStatus
from py_experimenter.experimenter import PyExperimenter

if TYPE_CHECKING:
    from py_experimenter.result_processor import ResultProcessor


def py_experimenter_evaluate(parameters: dict, result_processor: ResultProcessor, custom_config: dict):
    config = parameters["config"]
    cfg_dict = json.loads(config)

    job_id = os.environ["BENCHMARKING_JOB_ID"]

    result_processor.process_results({"slurm_job_id": job_id})

    dict_config = OmegaConf.create(cfg_dict)
    cfg_path = f"{job_id}_hydra_config.yaml"
    OmegaConf.save(config=dict_config, f=cfg_path)

    with open(f"{job_id}_pyexperimenter_id.txt", "w+") as f:
        f.write(str(result_processor.experiment_id))

    with open(f"{job_id}_problem_container.txt", "w+") as f:
        f.write(cfg_dict["benchmark_id"])

    with open(f"{job_id}_optimizer_container.txt", "w+") as f:
        f.write(cfg_dict["optimizer_container_id"])

    return ExperimentStatus.PAUSED


def main() -> None:
    slurm_job_id = os.environ["BENCHMARKING_JOB_ID"]
    experiment_configuration_file_path = "carps/container/py_experimenter.yaml"

    if os.path.exists("carps/container/credentials.yaml"):
        database_credential_file_path = "carps/container/credentials.yaml"
    else:
        database_credential_file_path = None

    experimenter = PyExperimenter(
        experiment_configuration_file_path=experiment_configuration_file_path,
        name="example_notebook",
        database_credential_file_path=database_credential_file_path,
        log_file=f"logs/{slurm_job_id}.log",
        use_ssh_tunnel=OmegaConf.load(experiment_configuration_file_path).PY_EXPERIMENTER.Database.use_ssh_tunnel,
    )

    experimenter.execute(py_experimenter_evaluate, max_experiments=1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import numpy as np


def pareto(costs: np.ndarray) -> np.ndarray:
    is_pareto = np.ones(costs.shape[0], dtype=bool)
    for i, c in enumerate(costs):
        if is_pareto[i]:
            is_pareto[is_pareto] = np.any(costs[is_pareto] < c, axis=1)
            is_pareto[i] = True
    return is_pareto



from __future__ import annotations

import hashlib
import json
import logging
from pathlib import Path

import hydra
from omegaconf import DictConfig, OmegaConf
from py_experimenter.exceptions import DatabaseConnectionError
from py_experimenter.experimenter import PyExperimenter
from carps.utils.loggingutils import CustomEncoder

logger = logging.getLogger("create experiments")


@hydra.main(config_path="../configs", config_name="base.yaml", version_base=None)  # type: ignore[misc]
def main(cfg: DictConfig) -> None:
    """Run optimizer on problem.

    Save trajectory and metadata to database.

    Parameters
    ----------
    cfg : DictConfig
        Global configuration.

    """
    cfg_dict = OmegaConf.to_container(cfg=cfg, resolve=True)

    experiment_configuration_file_path = (
        cfg.pyexperimenter_configuration_file_path or Path(__file__).parent / "py_experimenter.yaml"
    )

    database_credential_file_path = cfg.database_credential_file_path or Path(__file__).parent / "credentials.yaml"
    if database_credential_file_path is not None and not database_credential_file_path.exists():
        database_credential_file_path = None

    experimenter = PyExperimenter(
        experiment_configuration_file_path=experiment_configuration_file_path,
        name="carps",
        database_credential_file_path=database_credential_file_path,
        log_level=logging.INFO,
        use_ssh_tunnel=OmegaConf.load(experiment_configuration_file_path).PY_EXPERIMENTER.Database.use_ssh_tunnel,
    )

    cfg_json = OmegaConf.to_container(cfg, resolve=True)

    # This value will always be unique so it
    # disables duplicate checking when adding entries to the database.
    # Py_experimenter will add a creation date so the information
    # is not lost.
    if "timestamp" in cfg_json:
        del cfg_json["timestamp"]

    # Compute hash to efficiently compare configs.
    # In MySQL json objects are reordered to improve performance.
    # This means that there is no guarantee of the json strings
    # to be equal.
    cfg_str = json.dumps(cfg_json, cls=CustomEncoder)
    cfg_hash = hashlib.sha256(cfg_str.encode()).hexdigest()

    rows = [
        {
            "config": cfg_str,
            "config_hash": cfg_hash,
            "benchmark_id": cfg_dict["benchmark_id"],
            "problem_id": cfg_dict["problem_id"],
            "optimizer_id": cfg_dict["optimizer_id"],
            "optimizer_container_id": cfg_dict["optimizer_container_id"],
            "seed": cfg_dict["seed"],
            "n_trials": cfg_dict["task"]["n_trials"],
            "time_budget": cfg_dict["task"]["time_budget"],
        }
    ]

    column_names = list(experimenter.db_connector.database_configuration.keyfields.keys())

    exists = False

    try:
        existing_rows = experimenter.db_connector._get_existing_rows(column_names)

        # Check if experiment exists
        # Compare by hash only
        for e in existing_rows:
            if e["config_hash"] == cfg_hash:
                exists = True
                logger.info("Experiment not added to the database because config hash already exists!")
    except DatabaseConnectionError as e:
        if "1146" in e.args[0] or "no such table" in e.args[0]:
            logger.info("Database empty, will fill.:)")
        else:
            raise e

    if not exists:
        experimenter.fill_table_with_rows(rows)


if __name__ == "__main__":
    main()



"""MF-Prior-Bench Problem.

* HOW TO USE:
* ------------
* MF-Prior-Bench Documentation: https://automl.github.io/mf-prior-bench/latest/setup/
* 1. Install MF-Prior-Bench using `pip install mf-prior-bench`
* 2. Download the data for the respective benchmarks using
*    `python -m mfpbench download --benchmark <benchmark_name> --datadir <data_dir_path>`
*     where <benchmark_name> is one of ["pd1", "jahs"]
* 3. Install requirements from "./container_recipes/benchmarks/MFPBench/MFPBench_requirements.txt"
*    NOTE: JAHSBench is commented out in the requirements file due to compatibility issues
* 4. Test example 1 (smac20 multifidelity on PD1 imagenet_resnet_512 benchmark):
*    `python carps/run.py +optimizer/smac20=multifidelity +problem/MFPBench/pd1=imagenet_resnet_512
*     seed=1 task.n_trials=25 data_dir=<data_dir_path>`
*    Test example 2 (smac20 multifidelity on all available MFHartmann benchmarks):
*    `python carps/run.py +optimizer/smac20=multifidelity '+problem/MFPBench/mfh=glob(*)' 'seed=range(0, 10)' -m`
"""

from __future__ import annotations

import time
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

import mfpbench

from carps.benchmarks.problem import Problem
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import ConfigurationSpace

    from carps.loggers.abstract_logger import AbstractLogger

benchmarks_names = ["pd1", "jahs", "mfh"]

benchmarks = [
    # PD1
    "lm1b_transformer_2048",
    "translatewmt_xformer_64",
    "cifar100_wideresnet_2048",
    "imagenet_resnet_512",
    # MFHartmann
    "mfh3",
    "mfh6",
    "mfh3_terrible",
    "mfh3_bad",
    "mfh3_moderate",
    "mfh3_good",
    "mfh6_terrible",
    "mfh6_bad",
    "mfh6_moderate",
    "mfh6_good",
    # JAHSBench
    "jahs",  # NOTE: Untested, compatibility issues with some package versions
]

C = TypeVar("C", bound=mfpbench.Config)


class MFPBenchProblem(Problem):
    """MF-Prior-Bench Problem class."""

    def __init__(
        self,
        benchmark_name: str,
        metric: str | list[str],
        benchmark: str | None = None,
        budget_type: str | None = None,
        prior: str | Path | C | Mapping[str, Any] | None = None,
        perturb_prior: float | None = None,
        benchmark_kwargs: dict | None = None,
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        """Initialize a MF-Prior-Bench problem."""
        super().__init__(loggers)

        self.benchmark_name = benchmark_name
        self.budget_type = budget_type
        self.benchmark = benchmark
        self.metrics = metric
        self.prior = prior
        self.perturb_prior = perturb_prior
        assert self.benchmark_name in benchmarks_names, f"benchmark_name must be one of {benchmarks_names}"
        assert self.benchmark in benchmarks, f"benchmark '{benchmark}' must be one of {benchmarks}"

        if benchmark_kwargs is None:
            benchmark_kwargs = {}
        elif benchmark_kwargs.get("datadir") is not None:
            # Assumes that the data is stored in the following format:
            benchmark_kwargs["datadir"] = Path(benchmark_kwargs["datadir"]) / benchmark_name

        self._problem = mfpbench.get(
            name=benchmark,
            prior=self.prior,
            perturb_prior=self.perturb_prior,
            **benchmark_kwargs,
        )
        self._configspace = self._problem.space

    @property
    def configspace(self) -> ConfigurationSpace:
        """Return configuration space.

        Returns:
        -------
        ConfigurationSpace
            Configuration space.
        """
        return self._configspace

    def _evaluate(self, trial_info: TrialInfo) -> TrialValue:
        """Evaluate problem.

        Parameters
        ----------
        trial_info : TrialInfo
            Dataclass with configuration, seed, budget, instance.

        Returns:
        -------
        TrialValue
            Cost
        """
        configuration = trial_info.config
        start_time = time.time()
        result = self._problem.query(
            configuration.get_dictionary(),
            at=int(trial_info.budget) if trial_info.budget is not None else None,
        ).as_dict()
        end_time = time.time()

        ret = [result[metric] for metric in self.metrics]
        if len(ret) == 1:
            ret = ret[0]

        vt = 0.0
        if self.benchmark_name == "jahs":
            vt = result["runtime"]
        elif self.benchmark_name == "mfh":
            vt = result["fid_cost"]
        else:
            vt = result["train_cost"]

        return TrialValue(
            cost=ret,
            time=end_time - start_time,
            starttime=start_time,
            endtime=end_time,
            virtual_time=vt,
        )



from __future__ import annotations

import json
import os

from ConfigSpace.read_and_write import json as cs_json
from flask import Flask, request
from omegaconf import OmegaConf

from carps.utils.running import make_problem
from carps.utils.trials import TrialInfo
from carps.utils.loggingutils import CustomEncoder

if (job_id := os.environ["BENCHMARKING_JOB_ID"]) != "":
    cfg = OmegaConf.load(f"{job_id}_hydra_config.yaml")

problem = make_problem(cfg=cfg)

# TODO Check that problem container and problem match

app = Flask(__name__)
app.run()


@app.route("/configspace", methods=["GET"])
def _request_configspace() -> str:
    return json.dumps(cs_json.write(problem.configspace), cls=CustomEncoder)


@app.route("/evaluate", methods=["POST"])
def _request_evaluation() -> str:
    if request.is_json:
        trial_info = TrialInfo(**json.loads(request.get_json()))
        return json.dumps(problem.evaluate(trial_info).to_json(), cls=CustomEncoder)
    else:
        raise ValueError("Request is not JSON.")



from __future__ import annotations

import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import pkg_resources

if TYPE_CHECKING:
    from omegaconf import DictConfig


def _check(p: str | Path) -> None:
    """Check requirements specified in p.

    Parameters
    ----------
    p : str | Path
        Path to requirements. Txt file.

    Raises:
    ------
    RuntimeError
        When the requirement is not installed.

    Warning:
        When there is a version conflict.
    """
    p = Path(p)
    if p.is_file():
        with open(p) as file:
            requirements = file.readlines()
        requirements = [r.strip() for r in requirements]
        requirements = [r for r in requirements if "git+" not in r]

        try:
            pkg_resources.require(requirements)
        except pkg_resources.DistributionNotFound as error:
            error_msg = str(error)
            msg = (
                f"{error_msg}. Please install all necessary requirements with\n"
                f"\t>>>>>>>>>> pip install -r {p}\n"
                "You can also build an env for that specific combination, check `CARP-S/scripts/build_env(s).sh`."
            )
            raise RuntimeError(msg)
        except pkg_resources.VersionConflict as error:
            error_msg = str(error)
            msg = (
                f"Version Conflict: {error_msg} for {p}. Resolve manually. If it is about YAHPO and ConfigSpace, run the following "
                "to make YAHPO compatible with newest ConfigSpace:\n"
                f"\t>>>>>>>>>> python {p.parent.parent.parent.parent / 'scripts/patch_yahpo_configspace.py'}"
            )
            warnings.warn(msg)


def check_requirements(cfg: DictConfig) -> None:
    """Check whether requirements are satisified for benchmark and optimizer.

    Parameters
    ----------
    cfg : DictConfig
        Experiment configuration
    """
    p_base = Path(__file__).parent.parent.parent / "container_recipes"
    req_file_benchmark = p_base / "benchmarks" / cfg.benchmark_id / f"{cfg.benchmark_id}_requirements.txt"
    req_file_optimizer = (
        p_base / "optimizers" / cfg.optimizer_container_id / f"{cfg.optimizer_container_id}_requirements.txt"
    )

    _check(req_file_benchmark)
    _check(req_file_optimizer)



"""Nevergrad optimizer.

* Nevergrad Documentation: https://facebookresearch.github.io/nevergrad/
* List of Optimizers available:
* https://facebookresearch.github.io/nevergrad/optimizers_ref.html#optimizers
* Other Optimizers:
* Hyperopt: https://github.com/hyperopt/hyperopt,
* CMA-ES: https://github.com/CMA-ES/pycma,
* bayes_opt: https://github.com/bayesian-optimization/BayesianOptimization
* DE: https://facebookresearch.github.io/nevergrad/optimizers_ref.html#nevergrad.families.DifferentialEvolution
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import ConfigSpace.hyperparameters as CSH
import nevergrad as ng
import numpy as np
from ConfigSpace import Configuration, ConfigurationSpace

from carps.optimizers.optimizer import Optimizer
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from nevergrad.optimization.base import (
        ConfiguredOptimizer as ConfNGOptimizer,
        Optimizer as NGOptimizer,
    )
    from nevergrad.parametrization import parameter
    from omegaconf import DictConfig

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent


opt_list = sorted(ng.optimizers.registry.keys())
ext_opts = {
    "Hyperopt": ng.optimization.optimizerlib.NGOpt13,
    "CMA-ES": ng.optimization.optimizerlib.ParametrizedCMA,
    "bayes_opt": ng.optimization.optimizerlib.ParametrizedBO,
    "DE": ng.families.DifferentialEvolution,
    "EvolutionStrategy": ng.families.EvolutionStrategy,
}


def CS_to_nevergrad_space(hp: CSH.Hyperparameter) -> ng.p.Instrumentation:
    """Convert ConfigSpace to Nevergrad Parameter."""
    if isinstance(hp, CSH.FloatHyperparameter):
        if hp.log:
            return ng.p.Log(lower=hp.lower, upper=hp.upper)
        else:
            return ng.p.Scalar(lower=hp.lower, upper=hp.upper)
    elif isinstance(hp, CSH.IntegerHyperparameter):
        if hp.log:
            return ng.p.Log(lower=hp.lower, upper=hp.upper).set_integer_casting()
        else:
            return ng.p.Scalar(lower=hp.lower, upper=hp.upper).set_integer_casting()
    elif isinstance(hp, CSH.CategoricalHyperparameter):
        return ng.p.Choice(hp.choices)
    elif isinstance(hp, CSH.OrdinalHyperparameter):
        return ng.p.TransitionChoice(hp.sequence)
    elif isinstance(hp, CSH.Constant):
        return ng.p.Choice([hp.value])
    else:
        raise NotImplementedError(f"Unknown hyperparameter type: {hp.__class__.__name__}")


class NevergradOptimizer(Optimizer):
    """An optimizer that uses Nevergrad to optimize an objective function."""

    def __init__(
        self,
        problem: Problem,
        nevergrad_cfg: DictConfig,
        optimizer_cfg: DictConfig,
        task: Task,
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        super().__init__(problem, task, loggers)

        self.fidelity_enabled = False
        self.fidelity_type = None
        if self.task.is_multifidelity:
            self.fidelity_enabled = True
            self.fidelity_type: str = self.task.fidelity_type
        self.task = task
        self.configspace = problem.configspace
        self.ng_space = self.convert_configspace(self.configspace)
        self.nevergrad_cfg = nevergrad_cfg
        self.optimizer_cfg = optimizer_cfg
        if self.optimizer_cfg is None:
            self.optimizer_cfg = {}
        if self.nevergrad_cfg.optimizer_name not in opt_list and self.nevergrad_cfg.optimizer_name not in ext_opts:
            raise ValueError(f"Optimizer {self.nevergrad_cfg.optimizer_name} not found in Nevergrad!")

        self._solver: NGOptimizer | ConfNGOptimizer | None = None
        self.counter = 0
        self.history: dict[str, tuple[ng.p.Parameter, float | list[float] | None]] = {}

    def convert_configspace(self, configspace: ConfigurationSpace) -> ng.p.Parameter:
        """Convert ConfigSpace configuration space to search space from optimizer.

        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space from Problem.

        Returns:
        -------
        SearchSpace
            Optimizer's search space.
        """
        ng_param = ng.p.Dict()
        for hp in configspace.get_hyperparameters():
            ng_param[hp.name] = CS_to_nevergrad_space(hp)
        return ng_param

    def _setup_optimizer(self) -> NGOptimizer | ConfNGOptimizer:
        if self.nevergrad_cfg.optimizer_name in ext_opts:
            if self.nevergrad_cfg.optimizer_name == "Hyperopt":
                ng_opt = ext_opts[self.nevergrad_cfg.optimizer_name]
            else:
                ng_opt = ext_opts[self.nevergrad_cfg.optimizer_name](**self.optimizer_cfg)
            ng_opt = ng_opt(
                parametrization=self.ng_space,
                budget=self.task.n_trials,
                num_workers=self.task.n_workers,
            )
        else:
            ng_opt = ng.optimizers.registry[self.nevergrad_cfg.optimizer_name](
                parametrization=self.ng_space,
                budget=self.task.n_trials,
                num_workers=self.task.n_workers,
            )
        ng_opt.parametrization.random_state = np.random.RandomState(self.nevergrad_cfg.seed)
        return ng_opt

    def convert_to_trial(  # type: ignore[override]
        self,
        config: Configuration,
        name: str | None = None,
        seed: int | None = None,
        budget: float | None = None,
    ) -> TrialInfo:
        """Convert proposal from Nevergrad to TrialInfo.

        This ensures that the problem can be evaluated with a unified API.

        Parameters
        ----------
        config : Configuration
            Configuration from Nevergrad.
        name : str, optional
            Name of the trial, by default None
        seed : int, optional
            Seed of the trial, by default None

        Returns:
        -------
        TrialInfo
            Trial info containing configuration, budget, seed, instance.
        """
        return TrialInfo(
            config=config,
            name=name,
            seed=self.nevergrad_cfg.seed,
            budget=budget,
            instance=None,
        )

    def ask(self) -> TrialInfo:
        """Ask the optimizer for a new trial to evaluate.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Returns:
        -------
        TrialInfo
            trial info (config, seed, instance, budget)
        """
        config: parameter.Parameter = self.solver.ask()
        unique_name = f"{self.counter}_{config.value}_{self.nevergrad_cfg.seed}"
        self.history[unique_name] = (config, None)
        trial_info = self.convert_to_trial(
            config=Configuration(self.configspace, values=config.value, allow_inactive_with_values=True),
            name=unique_name,
            seed=self.nevergrad_cfg.seed,
            budget=None if not self.fidelity_enabled else self.task.max_budget,
        )
        self.counter += 1
        return trial_info

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        """Tell the optimizer a new trial.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Parameters
        ----------
        trial_info : TrialInfo
            trial info (config, seed, instance, budget)
        trial_value : TrialValue
            trial value (cost, time, ...)
        """
        unique_name = trial_info.name
        assert unique_name is not None
        assert unique_name in self.history
        assert self._solver is not None

        self.history[unique_name] = (self.history[unique_name][0], trial_value.cost)

        self.solver.tell(
            self.history[unique_name][0],
            trial_value.cost,
        )

    def get_current_incumbent(self) -> Incumbent:
        """Extract the incumbent config and cost. May only be available after a complete run.

        Returns:
        -------
        Incumbent: tuple[TrialInfo, TrialValue] | list[tuple[TrialInfo, TrialValue]] | None
            The incumbent configuration with associated cost.
        """
        incumbent = None
        cost = None
        unique_name = None
        if self.task.n_objectives > 1:
            configs = self.solver.pareto_front()
            costs = [param.losses.tolist() for param in configs]
            trial_info = [
                self.convert_to_trial(config=Configuration(self.configspace, values=config.value)) for config in configs
            ]
            trial_value = [TrialValue(cost=cost) for cost in costs]
            incumbent_tuple = list(zip(trial_info, trial_value, strict=False))
        else:
            for name, value in self.history.items():
                if incumbent is None or value[1] < cost:
                    incumbent = value[0].value
                    cost = value[1]
                    unique_name = name
            if cost is None:
                raise ValueError(f"Tried to get Incumbent without calling tell() for config {incumbent}!")
            trial_info = self.convert_to_trial(
                config=Configuration(self.configspace, values=incumbent, allow_inactive_with_values=True),
                name=unique_name,
                seed=self.nevergrad_cfg.seed,
            )
            trial_value = TrialValue(cost=cost)
            incumbent_tuple = (trial_info, trial_value)
        return incumbent_tuple



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from carps.optimizers.optimizer import Optimizer
from carps.utils.pareto_front import pareto
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import Configuration, ConfigurationSpace

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent, SearchSpace


class RandomSearchOptimizer(Optimizer):
    def __init__(
        self,
        problem: Problem,
        task: Task,
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        super().__init__(problem, task, loggers)

        self.configspace: ConfigurationSpace = self.problem.configspace
        self.history: list[tuple[TrialInfo, TrialValue]] = []
        self.is_multifidelity = task.is_multifidelity

        if hasattr(task, "n_objectives"):
            self.is_multiobjective = task.n_objectives > 1

    def convert_configspace(self, configspace: ConfigurationSpace) -> SearchSpace:
        return configspace

    def convert_to_trial(self, config: Configuration) -> TrialInfo:  # type: ignore[override]
        budget = None
        if self.is_multifidelity:
            budget = self.task.max_budget
            # budget = np.random.choice(np.linspace(self.task.min_budget, self.task.max_budget, 5))
        return TrialInfo(config=config, budget=budget)

    def ask(self) -> TrialInfo:
        config = self.problem.configspace.sample_configuration()
        return self.convert_to_trial(config=config)

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        self.history.append((trial_info, trial_value))

    def _setup_optimizer(self) -> None:
        return None

    def get_pareto_front(self) -> list[tuple[TrialInfo, TrialValue]]:
        """Return the pareto front for multi-objective optimization."""
        if self.task.is_multifidelity:
            max_budget = np.max([v[0].budget for v in self.history])
            results_on_highest_fidelity = np.array([v for v in self.history if v[0].budget == max_budget])
            costs = np.array([v[1].cost for v in results_on_highest_fidelity])
            # Determine pareto front of the trials run on max budget
            front = results_on_highest_fidelity[pareto(costs)]
        else:
            costs = np.array([v[1].cost for v in self.history])
            front = np.array(self.history)[pareto(costs)]
        return front.tolist()

    def get_current_incumbent(self) -> Incumbent:
        if self.task.n_objectives == 1:
            incumbent_tuple = min(self.history, key=lambda x: x[1].cost)
        else:
            incumbent_tuple = self.get_pareto_front()
        return incumbent_tuple



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from omegaconf import OmegaConf
from py_experimenter.experimenter import PyExperimenter

from carps.container.wrapper import ContainerizedProblemClient
from carps.loggers.database_logger import DatabaseLogger
from carps.loggers.file_logger import FileLogger
from carps.utils.running import make_optimizer

if TYPE_CHECKING:
    from py_experimenter.result_processor import ResultProcessor


def optimizer_experiment(parameters: dict, result_processor: ResultProcessor, custom_config: dict):
    loggers = [DatabaseLogger(result_processor), FileLogger()]
    problem = ContainerizedProblemClient(loggers=loggers)
    optimizer = make_optimizer(cfg=cfg, problem=problem)

    optimizer.run()


if (job_id := os.environ["BENCHMARKING_JOB_ID"]) != "":
    with open(f"{job_id}_pyexperimenter_id.txt") as f:
        experiment_id = int(f.read())

    cfg = OmegaConf.load(f"{job_id}_hydra_config.yaml")

    slurm_job_id = os.environ["BENCHMARKING_JOB_ID"]
    experiment_configuration_file_path = "carps/container/py_experimenter.yaml"

    if os.path.exists("carps/container/credentials.yaml"):
        database_credential_file = "carps/container/credentials.yaml"
    else:
        database_credential_file = None

    experimenter = PyExperimenter(
        experiment_configuration_file_path=experiment_configuration_file_path,
        name="example_notebook",
        database_credential_file_path=database_credential_file,
        log_file=f"logs/{slurm_job_id}.log",
        use_ssh_tunnel=OmegaConf.load(experiment_configuration_file_path).PY_EXPERIMENTER.Database.use_ssh_tunnel,
    )

    experimenter.unpause_experiment(experiment_id, optimizer_experiment)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from omegaconf import DictConfig, OmegaConf

from carps.optimizers.optimizer import Optimizer
from carps.utils.exceptions import AskAndTellNotSupportedError
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import Configuration, ConfigurationSpace
    from smac.facade.smac_ac_facade import SMAC4AC

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent


class NotSupportedError(Exception):
    pass


class SMAC314Optimizer(Optimizer):
    def __init__(
        self,
        problem: Problem,
        smac_cfg: DictConfig,
        task: Task,
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        super().__init__(problem, task, loggers)

        self.configspace = self.problem.configspace
        self.smac_cfg = smac_cfg
        self._solver: SMAC4AC | None = None

    def convert_configspace(self, configspace: ConfigurationSpace) -> ConfigurationSpace:
        """Convert configuration space from Problem to Optimizer.

        Here, we don't need to convert.

        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space from Problem.

        Returns:
        -------
        ConfigurationSpace
            Configuration space for Optimizer.
        """
        return configspace

    def convert_to_trial(  # type: ignore[override]
        self, config: Configuration, seed: int | None = None, budget: float | None = None, instance: str | None = None
    ) -> TrialInfo:
        """Convert proposal from SMAC to TrialInfo.

        Parameters
        ----------
        config : Configuration
            Configuration
        seed : int | None, optional
            Seed, by default None
        budget : float | None, optional
            Budget, by default None
        instance : str | None, optional
            Instance, by default None

        Returns:
        -------
        TrialInfo
            Trial info containing configuration, budget, seed, instance.
        """
        return TrialInfo(config=config, seed=seed, budget=budget, instance=instance)

    def target_function(
        self, config: Configuration, seed: int | None = None, budget: float | None = None, instance: str | None = None
    ) -> float | list[float]:
        """Target Function.

        Interface for the Problem.

        Parameters
        ----------
        config : Configuration
            Configuration
        seed : int | None, optional
            Seed, by default None
        budget : float | None, optional
            Budget, by default None
        instance : str | None, optional
            Instance, by default None

        Returns:
        -------
        float | list[float]
            Cost as float or list[float], depending on the number of objectives.
        """
        trial_info = self.convert_to_trial(config=config, seed=seed, budget=budget, instance=instance)
        trial_value = self.problem.evaluate(trial_info=trial_info)
        return trial_value.cost

    def _setup_optimizer(self) -> SMAC4AC:
        """Setup SMAC.

        Retrieve defaults and instantiate SMAC.

        Returns:
        -------
        SMAC4AC
            Instance of a SMAC facade.

        """
        from smac.facade.smac_ac_facade import SMAC4AC
        from smac.facade.smac_bb_facade import SMAC4BB
        from smac.facade.smac_hpo_facade import SMAC4HPO
        from smac.facade.smac_mf_facade import SMAC4MF
        from smac.scenario.scenario import Scenario

        if self.smac_cfg.scenario.n_workers > 1 and self.smac_cfg.optimization_type != "mf":
            raise NotSupportedError("SMAC 1.4 does not support parallel execution natively.")

        if self.smac_cfg.scenario.wallclock_limit is None:
            self.smac_cfg.scenario.wallclock_limit = np.inf

        # Instantiate Scenario
        scenario_kwargs = {
            "cs": self.configspace,
            "output_dir": None,
        }
        # We always expect scenario kwargs from the user
        _scenario_kwargs = OmegaConf.to_container(self.smac_cfg.scenario, resolve=True)
        scenario_kwargs.update(_scenario_kwargs)

        scenario = Scenario(scenario=scenario_kwargs)

        intensifier_kwargs: dict[Any, Any] = {}
        facade_kwargs: dict[Any, Any] = {}

        # Create facade
        if self.smac_cfg.optimization_type == "bb":
            facade_object = SMAC4BB
            intensifier_kwargs["maxR"] = self.smac_cfg.max_config_calls

        elif self.smac_cfg.optimization_type == "hpo":
            facade_object = SMAC4HPO
            intensifier_kwargs["maxR"] = self.smac_cfg.max_config_calls

        elif self.smac_cfg.optimization_type == "mf":
            facade_object = SMAC4MF

            n_seeds = self.smac_cfg.get("n_seeds", None)
            intensifier_kwargs["n_seeds"] = n_seeds
            intensifier_kwargs["initial_budget"] = self.smac_cfg.scenario.min_budget
            intensifier_kwargs["max_budget"] = self.smac_cfg.scenario.max_budget

            inc_selection = self.smac_cfg.incumbent_selection
            if inc_selection == "highest_observed_budget":
                inc_selection = "highest_budget"

            intensifier_kwargs["incumbent_selection"] = inc_selection
            facade_kwargs["n_jobs"] = self.smac_cfg.scenario.n_workers
        elif self.smac_cfg.optimization_type == "ac":
            facade_object = SMAC4AC
            intensifier_kwargs["maxR"] = self.smac_cfg.max_config_calls

        else:
            raise RuntimeError("Unknown optimization type.")

        if self.smac_cfg.intensifier is None:
            intensifier = None
        elif self.smac_cfg.intensifier == "successive_halving":
            from smac.intensification.successive_halving import SuccessiveHalving

            intensifier = SuccessiveHalving
        else:
            raise RuntimeError("Unsupported intensifier.")

        return facade_object(
            scenario=scenario,
            tae_runner=self.target_function,
            intensifier=intensifier,
            intensifier_kwargs=intensifier_kwargs,
            **facade_kwargs,
        )

    def ask(self) -> TrialInfo:
        raise AskAndTellNotSupportedError

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        raise AskAndTellNotSupportedError

    def _run(self) -> Incumbent:
        """Run SMAC on Problem."""
        incumbent = self.solver.optimize()  # noqa: F841
        return self.get_current_incumbent()

    def get_current_incumbent(self) -> Incumbent:
        trial_info = TrialInfo(config=self.solver.solver.incumbent)
        trial_value = TrialValue(cost=self.solver.get_runhistory().get_cost(self.solver.solver.incumbent))
        return (trial_info, trial_value)



from __future__ import annotations

import subprocess
from pathlib import Path
from typing import TYPE_CHECKING

import hydra
from hydra.core.hydra_config import HydraConfig

from carps.utils.loggingutils import get_logger, setup_logging

if TYPE_CHECKING:
    from omegaconf import DictConfig

setup_logging()
logger = get_logger(__file__)


@hydra.main(config_path="configs", config_name="base.yaml", version_base=None)  # type: ignore[misc]
def main(cfg: DictConfig) -> None:
    """Run optimizer on problem.

    Save trajectory and metadata to database.

    Parameters
    ----------
    cfg : DictConfig
        Global configuration.

    """
    hydra_cfg = HydraConfig.instance().get()
    overrides = hydra_cfg.overrides.task
    runcommand = f"python -m carps.run {' '.join(overrides)}"
    logger.info(f"Runcommand: `{runcommand}`")
    env_location = Path(cfg.conda_env_location) / cfg.conda_env_name
    logger.info(f"Selected environment: {env_location}")
    command = f"source ~/.bashrc; micromamba run -p {env_location} {runcommand}"
    logger.info(command)
    subprocess.run(command, shell=True, check=False)


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import numpy as np
import json

from rich.logging import RichHandler


def setup_logging() -> None:
    FORMAT = "%(message)s"
    logging.basicConfig(level=logging.INFO, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])


def get_logger(logger_name: str) -> logging.Logger:
    """Get the logger by name."""
    return logging.getLogger(logger_name)

class CustomEncoder(json.JSONEncoder):
    """
    - Serializes python/Numpy objects via customizing json encoder.
    - **Usage**
        - `json.dumps(python_dict, cls=EncodeFromNumpy)` to get json string.
        - `json.dump(*args, cls=EncodeFromNumpy)` to create a file.json.
    """
    def default(self, obj):
        if isinstance(obj, np.int64):
            return int(obj)
        elif isinstance(obj, np.float64):
            return float(obj)
        return super(CustomEncoder, self).default(obj)


from __future__ import annotations

from pathlib import Path

import fire
import pandas as pd
from omegaconf import OmegaConf

config_folder = Path(__file__).parent.parent / "configs"
config_folder_problem = config_folder / "problem"
config_folder_optimizer = config_folder / "optimizer"


def index_configs():
    """Index all problem and optimizer configs.

    Create `index.csv` containing the config filename `config_fn` and the
    """
    for key, path in zip(
        ["problem_id", "optimizer_id"], [config_folder_problem, config_folder_optimizer], strict=False
    ):
        paths = list(path.glob("**/*.yaml"))

        table = []
        for p in paths:
            cfg = OmegaConf.load(p)
            value = cfg.get(key)
            table.append(
                {
                    "config_fn": str(p),
                    key: value,
                }
            )
        table = pd.DataFrame(table)
        table.to_csv(path / "index.csv", index=False)


if __name__ == "__main__":
    fire.Fire(index_configs)



"""HPOBench problem class."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

from hpobench.benchmarks.ml.lr_benchmark import LRBenchmark
from hpobench.benchmarks.ml.nn_benchmark import NNBenchmark
from hpobench.benchmarks.ml.rf_benchmark import RandomForestBenchmark
from hpobench.benchmarks.ml.svm_benchmark import SVMBenchmark
from hpobench.benchmarks.ml.tabular_benchmark import TabularBenchmark
from hpobench.benchmarks.ml.xgboost_benchmark import XGBoostBenchmark
from omegaconf import ListConfig

from carps.benchmarks.problem import Problem
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import ConfigurationSpace
    from hpobench.container.client_abstract_benchmark import AbstractBenchmarkClient

    from carps.loggers.abstract_logger import AbstractLogger


class HPOBenchProblem(Problem):
    """HPOBench problem."""

    def __init__(
        self,
        seed: int,
        model: str | None = None,
        task_id: int | None = None,
        metric: str | list[str] = "function_value",
        problem: AbstractBenchmarkClient | None = None,
        budget_type: str | None = None,
        loggers: list[AbstractLogger] | None = None,
    ):
        """Initialize a HPOBench problem.

        Either specify model and task_id for an ML problem or problem.

        Parameters
        ----------
        model : str Model name.
        task_id : str Task ID, see https://arxiv.org/pdf/2109.06716.pdf, page 22.
        problem : AbstractBenchmarkClient
            Instantiated benchmark problem, e.g.
            `hpobench.container.benchmarks.surrogates.paramnet_benchmark.ParamNetAdultOnStepsBenchmark`.
        seed: int Random seed.
        budget_type : Optional[str] Budget type for the multifidelity setting.
                      Should be None for the blackbox setting.
        """
        super().__init__(loggers)

        self.budget_type = budget_type

        if problem is None and model is None and task_id is None:
            raise ValueError("Please specify either problem or model and task_id.")

        self._problem = (
            problem
            if problem
            else get_hpobench_problem(task_id=task_id, model=model, seed=seed, budget_type=self.budget_type)
        )
        if not isinstance(metric, list | ListConfig):
            metric = [metric]
        self.metric = metric
        self._configspace = self._problem.get_configuration_space(seed=seed)

    @property
    def configspace(self) -> ConfigurationSpace:
        """Return configuration space.

        Returns:
        -------
        ConfigurationSpace
            Configuration space.
        """
        return self._configspace

    def _evaluate(self, trial_info: TrialInfo) -> TrialValue:
        """Evaluate problem.

        Parameters
        ----------
        trial_info : TrialInfo
            Dataclass with configuration, seed, budget, instance.

        Returns:
        -------
        TrialValue
            Cost
        """
        configuration = trial_info.config
        starttime = time.time()

        if trial_info.budget is not None and self.budget_type is not None:
            budget_value = float(trial_info.budget) if self.budget_type == "subsample" else round(trial_info.budget)
            fidelity = {self.budget_type: budget_value}
        else:
            fidelity = None

        result_dict = self._problem.objective_function(
            configuration=configuration, fidelity=fidelity, rng=trial_info.seed
        )
        endtime = time.time()
        T = endtime - starttime
        virtual_time = result_dict["cost"] if trial_info.budget is None else 0.0
        costs = [result_dict[metric] for metric in self.metric]
        if len(costs) == 1:
            costs = costs[0]
        # function_value is 1 - accuracy on the validation set
        return TrialValue(
            cost=costs,
            time=T,
            starttime=starttime,
            endtime=endtime,
            virtual_time=virtual_time,
        )


def get_hpobench_problem(model: str, task_id: int, seed: int, budget_type: str | None = None) -> Any:
    """Get HPOBench problem.

    Parameters
    ----------
        model : str Model name.
        task_id : str Task ID, see https://arxiv.org/pdf/2109.06716.pdf, page 22.
        seed: int Random seed.
        budget_type : Optional[str] Budget type for the multifidelity setting.
                      Should be None for the blackbox setting.

    Returns:
    -------
    Any
        Target function.
    """
    common_args = {"rng": seed, "task_id": task_id}
    if model == "lr":
        problem = TabularBenchmark(model="lr", **common_args) if budget_type is None else LRBenchmark(**common_args)
    elif model == "nn":
        problem = TabularBenchmark(model="nn", **common_args) if budget_type is None else NNBenchmark(**common_args)
    elif model == "rf":
        problem = (
            TabularBenchmark(model="rf", **common_args) if budget_type is None else RandomForestBenchmark(**common_args)
        )
    elif model == "svm":
        problem = TabularBenchmark(model="svm", **common_args) if budget_type is None else SVMBenchmark(**common_args)
    elif model == "xgboost":
        problem = (
            TabularBenchmark(model="xgb", **common_args) if budget_type is None else XGBoostBenchmark(**common_args)
        )
    else:
        raise ValueError(f"Unknown model {model} for HPOBench.")

    return problem



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from carps.utils.trials import TrialInfo, TrialValue
    from carps.utils.types import Incumbent


class AbstractLogger(ABC):
    def __init__(self) -> None:
        """AbstractLogger.

        Intercepts the trial info and value during evaluate.
        """

    @abstractmethod
    def log_trial(
        self, n_trials: float, trial_info: TrialInfo, trial_value: TrialValue, n_function_calls: int | None = None
    ) -> None:
        """Log the trial.

        Parameters
        ----------
        n_trials : float
            The number of trials that have been run so far.
            For the case of multi-fidelity, a full trial
            is a configuration evaluated on the maximum budget and
            the counter is increased by `budget/max_budget` instead
            of 1.
        trial_info : TrialInfo
            The trial info.
        trial_value : TrialValue
            The trial value.
        n_function_calls: int | None, default None
            The number of target function calls, no matter the budget.
        """
        raise NotImplementedError

    @abstractmethod
    def log_incumbent(self, n_trials: int, incumbent: Incumbent) -> None:
        """Log the incumbents.

        Parameters
        ----------
        n_trials : int
            The number of trials that have been run so far.
        incumbent : Incumbent
            The incumbent (or multiple incumbents).
        """
        raise NotImplementedError

    @abstractmethod
    def log_arbitrary(self, data: dict, entity: str) -> None:
        """Log arbitrary data.

        Parameters
        ----------
        entity : str
            The entity to which to log (e.g. filename, table name).
        data : dict
            The data to log.
        """
        raise NotImplementedError



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from hydra.utils import instantiate
from omegaconf import DictConfig, OmegaConf
from rich import (
    inspect,
    print as printr,
)

from carps.utils.exceptions import NotSupportedError

if TYPE_CHECKING:
    from py_experimenter.result_processor import ResultProcessor

    from carps.benchmarks.problem import Problem
    from carps.optimizers.optimizer import Optimizer


def make_problem(cfg: DictConfig, result_processor: ResultProcessor | None = None) -> Problem:
    """Make Problem.

    Parameters
    ----------
    cfg : DictConfig
        Global configuration.
    result_processor : ResultProcessor
        Py experimenter result processor, important for logging. Necessary to
        instantiate database logger.

    Returns:
    -------
    Problem
        Target problem.
    """
    problem_cfg = cfg.problem
    loggers = []
    if "loggers" in cfg:
        for logger in cfg.loggers:
            if "DatabaseLogger" in logger._target_:
                kwargs = {"result_processor": result_processor}
            elif "FileLogger" in logger._target_:
                kwargs = {"directory": cfg.outdir}
            else:
                kwargs = {}
            logger = instantiate(logger)(**kwargs)
            loggers.append(logger)
    return instantiate(problem_cfg, loggers=loggers)


def make_optimizer(cfg: DictConfig, problem: Problem) -> Optimizer:
    """Make Optimizer.

    Parameters
    ----------
    loggers : list[AbstractLogger]
        List of loggers to use.
    cfg : DictConfig
        Global configuration
    problem : Problem
        Target problem

    Returns:
    -------
    Optimizer
        Instantiated optimizer.
    """
    optimizer_cfg = cfg.optimizer
    optimizer = instantiate(optimizer_cfg)(problem=problem, task=cfg.task, loggers=problem.loggers)
    if "optimizer_wrappers" in cfg:
        for wrapper in cfg.optimizer_wrappers:
            optimizer = wrapper(optimizer)
    return optimizer


def optimize(cfg: DictConfig, result_processor: ResultProcessor | None = None) -> None:
    """Run optimizer on problem.

    Save trajectory and metadata to database.

    Parameters
    ----------
    cfg : DictConfig
        Global configuration.
    result_processor : ResultProcessor
        Py experimenter result processor, important for logging. Necessary to
        instantiate database logger.

    """
    os.environ["HYDRA_FULL_ERROR"] = "1"

    cfg_dict = OmegaConf.to_container(cfg=cfg, resolve=True)
    printr(cfg_dict)
    # hydra_cfg = HydraConfig.instance().get()
    # printr(hydra_cfg.run.dir)

    problem = make_problem(cfg=cfg, result_processor=result_processor)
    inspect(problem)

    optimizer = make_optimizer(cfg=cfg, problem=problem)
    inspect(optimizer)

    try:
        inc_tuple = optimizer.run()
        printr("Solution found: ", inc_tuple)
    except NotSupportedError:
        print("Not supported. Skipping.")
    except Exception as e:
        print("Something went wrong:")
        print(e)
        raise e



from __future__ import annotations

from dataclasses import asdict
from typing import TYPE_CHECKING, Any

from hydra.utils import get_class
from omegaconf import DictConfig, OmegaConf
from rich import print as printr

# from git import Repo
# from smac.callback.metadata_callback import MetadataCallback
from smac.multi_objective.parego import ParEGO
from smac.runhistory import (
    TrialInfo as SmacTrialInfo,
    TrialKey as SmacTrialKey,
    TrialValue as SmacTrialValue,
)
from smac.scenario import Scenario

from carps.optimizers.optimizer import Optimizer
from carps.utils.loggingutils import setup_logging
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import Configuration, ConfigurationSpace
    from smac.facade.abstract_facade import AbstractFacade

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent

setup_logging()


def maybe_inst_add_scenario(smac_kwargs: dict[str, Any], key: str, scenario: Scenario) -> dict[str, Any]:
    if key in smac_kwargs:
        smac_kwargs[key] = smac_kwargs[key](scenario=scenario)
    return smac_kwargs


class SMAC3Optimizer(Optimizer):
    def __init__(
        self,
        problem: Problem,
        smac_cfg: DictConfig,
        task: Task,
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        super().__init__(problem, task, loggers)

        self.configspace = self.problem.configspace
        self.smac_cfg = smac_cfg
        self._solver: AbstractFacade | None = None
        self._cb_on_start_called: bool = False

    def convert_configspace(self, configspace: ConfigurationSpace) -> ConfigurationSpace:
        """Convert configuration space from Problem to Optimizer.

        Here, we don't need to convert.

        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space from Problem.

        Returns:
        -------
        ConfigurationSpace
            Configuration space for Optimizer.
        """
        return configspace

    def convert_to_trial(  # type: ignore[override]
        self, config: Configuration, seed: int | None = None, budget: float | None = None, instance: str | None = None
    ) -> TrialInfo:
        """Convert proposal from SMAC to TrialInfo.

        Parameters
        ----------
        config : Configuration
            Configuration
        seed : int | None, optional
            Seed, by default None
        budget : float | None, optional
            Budget, by default None
        instance : str | None, optional
            Instance, by default None

        Returns:
        -------
        TrialInfo
            Trial info containing configuration, budget, seed, instance.
        """
        return TrialInfo(config=config, seed=seed, budget=budget, instance=instance)

    def target_function(
        self, config: Configuration, seed: int | None = None, budget: float | None = None, instance: str | None = None
    ) -> float | list[float]:
        """Target Function.

        Interface for the Problem.

        Parameters
        ----------
        config : Configuration
            Configuration
        seed : int | None, optional
            Seed, by default None
        budget : float | None, optional
            Budget, by default None
        instance : str | None, optional
            Instance, by default None

        Returns:
        -------
        float | list[float]
            Cost as float or list[float], depending on the number of objectives.
        """
        trial_info = self.convert_to_trial(config=config, seed=seed, budget=budget, instance=instance)
        trial_value = self.problem.evaluate(trial_info=trial_info)
        return trial_value.cost

    def _setup_optimizer(self) -> AbstractFacade:
        """Setup SMAC.

        Retrieve defaults and instantiate SMAC.

        Returns:
        -------
        SMAC4AC
            Instance of a SMAC facade.

        """
        # repo = Repo(".", search_parent_directories=True)
        # metadata_callback = MetadataCallback(
        #     repository=repo.working_tree_dir.split("/")[-1],
        #     branch=str(repo.active_branch),
        #     commit=str(repo.head.commit),
        #     command=" ".join([sys.argv[0][len(repo.working_tree_dir) + 1 :]] + sys.argv[1:]),
        #     additional_information={
        #         "cfg": OmegaConf.to_yaml(cfg=cfg),
        #     },
        # )

        # Select SMAC Facade
        smac_class = get_class(self.smac_cfg.smac_class)

        if smac_class == get_class("smac.facade.multi_fidelity_facade.MultiFidelityFacade"):
            self.fidelity_enabled = True

        # Setup other SMAC kwargs
        smac_kwargs = {}
        if self.smac_cfg.smac_kwargs is not None:
            smac_kwargs = OmegaConf.to_container(self.smac_cfg.smac_kwargs, resolve=True, enum_to_str=True)

        # Instantiate Scenario
        scenario_kwargs = {
            "configspace": self.configspace,
            # output_directory=Path(self.config.hydra.sweep.dir)
            # / "smac3_output",  # output directory is automatically set via config file
        }
        # We always expect scenario kwargs from the user
        _scenario_kwargs = OmegaConf.to_container(self.smac_cfg.scenario, resolve=True)
        scenario_kwargs.update(_scenario_kwargs)

        scenario = Scenario(**scenario_kwargs)

        smac_kwargs["scenario"] = scenario

        # Convert callbacks to list if necessary
        # Callbacks can come as a dict due to impossible hydra composition of
        # lists.
        if "callbacks" not in smac_kwargs:
            smac_kwargs["callbacks"] = []
        elif "callbacks" in smac_kwargs and type(smac_kwargs["callbacks"]) == dict:
            smac_kwargs["callbacks"] = list(smac_kwargs["callbacks"].values())
        elif "callbacks" in smac_kwargs and type(smac_kwargs["callbacks"]) == list:
            pass

        # If we have a custom intensifier we need to instantiate ourselves
        # because the helper methods in the facades expect a scenario.
        smac_kwargs = maybe_inst_add_scenario(smac_kwargs, "intensifier", scenario)

        if "acquisition_function" in smac_kwargs and "acquisition_maximizer" in smac_kwargs:
            if "acquisition_maximizer" in smac_kwargs:
                smac_kwargs["acquisition_maximizer"] = smac_kwargs["acquisition_maximizer"](
                    configspace=self.configspace, acquisition_function=smac_kwargs["acquisition_function"]
                )
                if hasattr(smac_kwargs["acquisition_maximizer"], "selector") and hasattr(
                    smac_kwargs["acquisition_maximizer"].selector, "expl2callback"
                ):
                    smac_kwargs["callbacks"].append(smac_kwargs["acquisition_maximizer"].selector.expl2callback)

        smac_kwargs = maybe_inst_add_scenario(smac_kwargs, "config_selector", scenario)
        smac_kwargs = maybe_inst_add_scenario(smac_kwargs, "initial_design", scenario)
        smac_kwargs = maybe_inst_add_scenario(smac_kwargs, "multi_objective_algorithm", scenario)

        printr(smac_class, smac_kwargs)

        smac = smac_class(
            target_function=self.target_function,
            **smac_kwargs,
        )
        printr(smac)

        return smac

    def ask(self) -> TrialInfo:
        """Ask the optimizer for a new trial to evaluate.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Returns:
        -------
        TrialInfo
            trial info (config, seed, instance, budget)
        """
        if not self._cb_on_start_called:
            self._cb_on_start_called = True
            for callback in self.solver.optimizer._callbacks:
                callback.on_start(self.solver.optimizer)
        smac_trial_info = self.solver.ask()
        return TrialInfo(
            config=smac_trial_info.config,
            instance=smac_trial_info.instance,
            budget=smac_trial_info.budget,
            seed=smac_trial_info.seed,
            name=None,
            checkpoint=None,
        )

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        """Tell the optimizer a new trial.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Parameters
        ----------
        trial_value : TrialValue
            trial value (cost, time, ...)
        """
        smac_trial_info = SmacTrialInfo(
            config=trial_info.config, instance=trial_info.instance, budget=trial_info.budget, seed=trial_info.seed
        )
        additional_info = trial_value.additional_info
        if self.task.n_objectives > 1:
            # Save costs for multi-objective because SMAC might scalarize the cost
            additional_info["cost"] = trial_value.cost
        smac_trial_value = SmacTrialValue(
            cost=trial_value.cost,
            time=trial_value.time,
            status=trial_value.status,
            starttime=trial_value.starttime,
            endtime=trial_value.endtime,
            additional_info=additional_info,
        )
        self.solver.tell(info=smac_trial_info, value=smac_trial_value)

    def get_current_incumbent(self) -> Incumbent:
        if self.solver.scenario.count_objectives() == 1:
            inc = self.solver.intensifier.get_incumbent()
            cost = self.solver.runhistory.get_cost(config=inc)
            trial_info = TrialInfo(config=inc)
            trial_value = TrialValue(cost=cost)
            incumbent_tuple = (trial_info, trial_value)
        else:
            if (mo := self.solver._runhistory_encoder.multi_objective_algorithm) is not None:
                if isinstance(mo, ParEGO) and mo._theta is None:
                    # Initialize weights of ParEGO
                    mo.update_on_iteration_start()
            incs = self.solver.intensifier.get_incumbents()
            tis = [
                self.solver.runhistory.get_trials(c)[0] for c in incs
            ]  # first trial because only one budget #TODO update for momf
            tks = [
                SmacTrialKey(
                    config_id=self.solver.runhistory.get_config_id(ti.config),
                    instance=ti.instance,
                    seed=ti.seed,
                    budget=ti.budget,
                )
                for ti in tis
            ]
            tvs = [self.solver.runhistory[k] for k in tks]
            tvs = [TrialValue(**asdict(tv)) for tv in tvs]
            incumbent_tuple = list(zip(tis, tvs, strict=False))

        return incumbent_tuple



from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
from typing import TYPE_CHECKING, Any

from dataclasses_json import dataclass_json

if TYPE_CHECKING:
    from ConfigSpace import Configuration


class StatusType(IntEnum):
    """Class to define status types of configs."""

    RUNNING = 0  # In case a job was submitted, but it has not finished.
    SUCCESS = 1
    CRASHED = 2
    TIMEOUT = 3
    MEMORYOUT = 4


@dataclass_json
@dataclass(frozen=True)
class TrialInfo:
    """Information about a trial.

    Parameters
    ----------
    config : Configuration
    instance : int | None, defaults to None, length
    seed : int | None, defaults to None
    budget : float | None, defaults to None
    normalized_budget: float | None, defaults to None
        The budget normalized by max_budget. Mostly used for logging purposes.
    name: str | None, defaults to None, arbitrary information, length 100
    checkpoint: str | None, defaults to None, checkpoint path, length 250

    The length of the strings depends on the setting for the database.
    """

    config: Configuration
    instance: int | None = None
    seed: int | None = None
    budget: float | None = None
    normalized_budget: float | None = None
    name: str | None = None
    checkpoint: str | None = None


@dataclass_json
@dataclass(frozen=True)
class TrialValue:
    """Values of a trial.

    Parameters
    ----------
    cost : float | list[float]
    time : float, defaults to 0.0
    virtual_time : float, defaults to 0.0
    status : StatusType, defaults to StatusType.SUCCESS
    starttime : float, defaults to 0.0
    endtime : float, defaults to 0.0
    additional_info : dict[str, Any], defaults to {}
    """

    cost: float | list[float]
    time: float = 0.0
    virtual_time: float = 0.0
    status: StatusType = StatusType.SUCCESS
    starttime: float = 0.0
    endtime: float = 0.0
    additional_info: dict[str, Any] = field(default_factory=dict)



"""Scikit-Optimize.

* Scikit-Optimize optimizer for CARPS.
* NOTE: Set to always minimize the cost
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import ConfigSpace as CS
import ConfigSpace.hyperparameters as CSH
import numpy as np
import skopt
from skopt.space.space import Categorical, Integer, Real, Space

from carps.optimizers.optimizer import Optimizer
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from omegaconf import DictConfig

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent


base_estimators = ["GP", "RF", "ET", "GBRT"]
acq_funcs = ["LCB", "EI", "PI", "gp_hedge"]
acq_optimizers = ["sampling", "lbfgs", "auto"]


def CS_to_skopt_space(hp: CSH.Hyperparameter) -> Space:
    """Convert ConfigSpace hyperparameter to skopt Space."""
    if isinstance(hp, CSH.FloatHyperparameter):
        if hp.log:
            return Real(hp.lower, hp.upper, name=hp.name, prior="log-uniform")
        else:
            return Real(hp.lower, hp.upper, name=hp.name)
    elif isinstance(hp, CSH.IntegerHyperparameter):
        if hp.log:
            return Integer(hp.lower, hp.upper, name=hp.name, prior="log-uniform")
        else:
            return Integer(hp.lower, hp.upper, name=hp.name)
    elif isinstance(hp, CSH.CategoricalHyperparameter):
        weights = None
        if hp.weights is not None:
            weights = np.asarray(hp.weights) / np.sum(hp.weights)
        return Categorical(hp.choices, name=hp.name, prior=weights)
    elif isinstance(hp, CSH.OrdinalHyperparameter):
        return Categorical(list(hp.sequence), name=hp.name)
    elif isinstance(hp, CSH.Constant):
        return Categorical([hp.value], name=hp.name)
    else:
        raise NotImplementedError(f"Unknown hyperparameter type: {hp.__class__.__name__}")


class SkoptOptimizer(Optimizer):
    """An optimizer that uses Scikit-Optimize to optimize an objective function."""

    def __init__(
        self,
        problem: Problem,
        skopt_cfg: DictConfig,
        task: Task,
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        """Initialize a Scikit-Optimize optimizer."""
        super().__init__(problem, task, loggers)

        self.fidelity_enabled = False

        self.configspace = problem.configspace

        self.skopt_space = self.convert_configspace(self.configspace)
        self._solver: skopt.optimizer.Optimizer | None = None

        self.skopt_cfg = skopt_cfg
        assert self.skopt_cfg.base_estimator in base_estimators
        assert self.skopt_cfg.acq_func in acq_funcs
        assert self.skopt_cfg.acq_optimizer in acq_optimizers

    def convert_configspace(self, configspace: CS.ConfigurationSpace) -> list[Space]:
        """Convert ConfigSpace configuration space to Scikit-Optimize Space.

        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space from Problem.

        Returns:
        -------
        list[Space]
            Scikit-Optimize Space.
        """
        space = []
        for hp in configspace.get_hyperparameters():
            space.append(CS_to_skopt_space(hp))
        return space

    def convert_to_trial(  # type: ignore[override]
        self, config: list[Any]
    ) -> TrialInfo:
        """Convert proposal by Scikit-Optimize to TrialInfo.

        This ensures that the problem can be evaluated with a unified API.

        Returns:
        -------
        TrialInfo
            Trial info containing configuration, budget, seed, instance.
        """
        configuration = CS.Configuration(
            configuration_space=self.configspace,
            values={hp.name: value for hp, value in zip(self.configspace.get_hyperparameters(), config, strict=False)},
            allow_inactive_with_values=True,
        )
        assert list(configuration.keys()) == list(self.configspace.get_hyperparameter_names())
        assert list(configuration.keys()) == [hp.name for hp in self.skopt_space]
        return TrialInfo(
            config=configuration,
            seed=self.skopt_cfg.random_state,
            budget=None,
            instance=None,
        )

    def _setup_optimizer(self) -> skopt.optimizer.Optimizer:
        if self.skopt_cfg is None:
            self.skopt_cfg = {}
        return skopt.optimizer.Optimizer(dimensions=self.skopt_space, **self.skopt_cfg)

    def ask(self) -> TrialInfo:
        """Ask the optimizer for a new trial to evaluate.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Returns:
        -------
        TrialInfo
            trial info (config, seed, instance, budget)
        """
        config = self.solver.ask()
        return self.convert_to_trial(config)

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        """Tell the optimizer a new trial.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Parameters
        ----------
        trial_info : TrialInfo
            trial info (config, seed, instance, budget)
        trial_value : TrialValue
            trial value (cost, time, ...)
        """
        _ = self.solver.tell(list(trial_info.config.values()), trial_value.cost)

    def get_current_incumbent(self) -> Incumbent:
        """Extract the incumbent config and cost. May only be available after a complete run.

        Returns:
        -------
        Incumbent: tuple[TrialInfo, TrialValue] | list[tuple[TrialInfo, TrialValue]] | None
            The incumbent configuration with associated cost.
        """
        best_result = self.solver.get_result()
        best_config = self.convert_to_trial(best_result.x)
        best_cost = TrialValue(cost=best_result.fun)
        return (best_config, best_cost)



from __future__ import annotations

import json
import multiprocessing
from collections.abc import Callable, Iterable
from dataclasses import asdict
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

import fire
import numpy as np
import pandas as pd
from ConfigSpace import Configuration
from hydra.core.utils import setup_globals
from omegaconf import DictConfig, ListConfig, OmegaConf

from carps.utils.loggingutils import get_logger, setup_logging
from carps.utils.task import Task
from carps.utils.trials import TrialInfo

if TYPE_CHECKING:
    from carps.benchmarks.problem import Problem

setup_logging()
logger = get_logger(__file__)


setup_globals()


def glob_trial_logs(p: str) -> list[str]:
    return list(Path(p).glob("**/trial_logs.jsonl"))


def get_run_dirs(outdir: str):
    opt_paths = list(Path(outdir).glob("*/*"))
    with multiprocessing.Pool() as pool:
        triallog_files = pool.map(glob_trial_logs, opt_paths)
    triallog_files = np.concatenate(triallog_files)
    return [f.parent for f in triallog_files]


def annotate_with_cfg(
    df: pd.DataFrame,
    cfg: DictConfig,
    config_keys: list[str],
    config_keys_forbidden: list[str] | None = None,
) -> pd.DataFrame:
    if config_keys_forbidden is None:
        config_keys_forbidden = []
    cfg_resolved = OmegaConf.to_container(cfg, resolve=True)
    flat_cfg = pd.json_normalize(cfg_resolved, sep=".").iloc[0].to_dict()  # type: ignore
    for k, v in flat_cfg.items():
        if np.any([k.startswith(c) for c in config_keys]) and not np.any([c in k for c in config_keys_forbidden]):
            if isinstance(v, list | ListConfig):
                v = [v] * len(df)
            df[k] = v
    return df


def get_Y(X: np.ndarray, problem: Problem) -> np.ndarray:
    return np.array(
        [
            (
                problem.evaluate(
                    trial_info=TrialInfo(config=Configuration(configuration_space=problem.configspace, vector=x))
                ).cost
            )
            for x in X
        ]
    )


def join_df(df1: pd.DataFrame, df2: pd.DataFrame, on: str = "n_trials") -> pd.DataFrame:
    df1.set_index(on)
    return df1.join(df2.set_index(on), on=on)


def load_log(rundir: str | Path, log_fn: str = "trial_logs.jsonl") -> pd.DataFrame:
    df = read_trial_log(rundir, log_fn=log_fn)
    if df is None:
        # raise NotImplementedError("No idea what should happen here!?")
        return pd.DataFrame()

    cfg = load_cfg(rundir)
    if cfg is not None:
        config_fn = str(Path(rundir) / ".hydra/config.yaml")
        cfg_str = OmegaConf.to_yaml(cfg=cfg)
        df["cfg_fn"] = config_fn
        df["cfg_str"] = [(config_fn, cfg_str)] * len(df)

        config_keys = ["benchmark", "problem", "seed", "optimizer_id", "task"]
        config_keys_forbidden = ["_target_", "_partial_"]
        df = annotate_with_cfg(df=df, cfg=cfg, config_keys=config_keys, config_keys_forbidden=config_keys_forbidden)
        # df = maybe_add_bandit_log(df, rundir, n_initial_design=cfg.task.n_initial_design)
    else:
        config_fn = "no_hydra_config"
        cfg_str = ""
        df["cfg_fn"] = config_fn
        df["cfg_str"] = [(config_fn, cfg_str)] * len(df)

    if "problem.function.seed" in df:
        df = df.drop(columns=["problem.function.seed"])

    if "problem.function.dim" in df:
        df = df.rename(columns={"problem.function.dim": "dim"})

    return process_logs(df)


T = TypeVar("T")
R = TypeVar("R")


def map_multiprocessing(
    task_function: Callable[[T], R],
    task_params: Iterable[T],
    n_processes: int | None = None,
) -> list[R]:
    with multiprocessing.Pool(processes=n_processes) as pool:
        return pool.map(task_function, task_params)


def read_jsonl_content(filename: str | Path) -> pd.DataFrame:
    with open(filename) as file:
        content = [json.loads(l) for l in file.readlines()]  # noqa: E741
    return pd.DataFrame(content)


def read_trial_log(rundir: str | Path, log_fn: str = "trial_logs.jsonl") -> pd.DataFrame | None:
    path = Path(rundir) / log_fn
    if not path.exists():
        return None

    df = read_jsonl_content(path)
    df = normalize_drop(df, "trial_info", rename_columns=True, sep="__")
    return normalize_drop(df, "trial_value", rename_columns=True, sep="__")
    # df = df.drop(columns=["trial_info__instance", "trial_info__budget", "trial_value__time", "trial_value__status", "trial_value__starttime", "trial_value__endtime"])


def load_cfg(rundir: str | Path) -> DictConfig | None:
    config_fn = Path(rundir) / ".hydra/config.yaml"
    if not config_fn.exists():
        return None

    return OmegaConf.load(config_fn)  # type: ignore


def normalize_drop(df: pd.DataFrame, key: str, rename_columns: bool = False, sep: str = ".") -> pd.DataFrame:
    """Normalize columns containing dicts.

    Parameters
    ----------
    df : pd.DataFrame
        Source data frame
    key : str
        Column name containing dicts
    rename_columns : bool, default False
        If true, rename columns to key.dict_key, if false, use dict keys as column names.
    sep : str, default .
        If rename columns: pattern for column is <key><sep><value>

    Returns:
    --------
    pd.DataFrame
        Flat data frame
    """
    df_tmp = pd.DataFrame(df[key].tolist())
    if "additional_info" in df_tmp and df_tmp["additional_info"].iloc[0] == {}:
        del df_tmp["additional_info"]
    if rename_columns:
        df_tmp = df_tmp.rename(columns={c: f"{key}{sep}{c}" for c in df_tmp.columns})
    return pd.concat([df.drop(key, axis=1), df_tmp], axis=1)


def maybe_add_n_trials(df: pd.DataFrame, n_initial_design: int, counter_key: str = "n_calls") -> pd.DataFrame:
    if "n_trials" not in df:
        df["n_trials"] = df[counter_key] + n_initial_design  # n_trials is 1-based
    return df


def add_scenario_type(logs: pd.DataFrame, task_prefix: str = "task.") -> pd.DataFrame:
    def determine_scenario_type(x: pd.Series) -> str:
        if x[task_prefix + "is_multifidelity"] is False and x[task_prefix + "is_multiobjective"] is False:
            scenario = "blackbox"
        elif x[task_prefix + "is_multifidelity"] is True and x[task_prefix + "is_multiobjective"] is False:
            scenario = "multi-fidelity"
        elif x[task_prefix + "is_multifidelity"] is False and x[task_prefix + "is_multiobjective"] is True:
            scenario = "multi-objective"
        elif x[task_prefix + "is_multifidelity"] is True and x[task_prefix + "is_multiobjective"] is True:
            scenario = "multi-fidelity-objective"
        elif np.isnan(x[task_prefix + "is_multifidelity"]) or np.isnan(x[task_prefix + "is_multiobjective"]):
            scenario = "blackbox"
        else:
            print(
                x["problem_id"],
                x["optimizer_id"],
                x["seed"],
                x[task_prefix + "is_multifidelity"],
                type(x[task_prefix + "is_multifidelity"]),
            )
            raise ValueError("Unknown scenario")
        return scenario

    logs["scenario"] = logs.apply(determine_scenario_type, axis=1)
    return logs


def maybe_postadd_task(logs: pd.DataFrame, overwrite: bool = False) -> pd.DataFrame:
    index_fn = Path(__file__).parent.parent / "configs/problem/index.csv"
    if not index_fn.is_file():
        raise ValueError("Problem ids have not been indexed. Run `python -m carps.utils.index_configs`.")
    problem_index = pd.read_csv(index_fn)

    def load_task_cfg(problem_id: str) -> DictConfig:
        subset = problem_index["config_fn"][problem_index["problem_id"] == problem_id]
        if len(subset) == 0:
            raise ValueError(
                f"Can't find config_fn for {problem_id}. Maybe the index is old. Run `python -m carps.utils.index_configs` to refresh."
            )
        config_fn = subset.iloc[0]
        if not Path(config_fn).is_file():
            raise ValueError(
                f"Can't find config_fn for {problem_id}. Maybe the index is old. Run `python -m carps.utils.index_configs` to refresh."
            )
        cfg = OmegaConf.load(config_fn)
        return cfg.task

    new_logs = []
    for gid, gdf in logs.groupby(by="problem_id"):
        task_cfg = load_task_cfg(problem_id=gid)
        task_columns = [c for c in gdf.columns if c.startswith("task.")]
        if overwrite:
            task_dict = asdict(Task(**task_cfg))
            task_columns = ["task." + k for k in task_dict]

        for c in task_columns:
            key = c.split(".")[1]
            # print(task_cfg, c)
            # print(gdf[c].explode().unique())
            if overwrite or np.nan in gdf[c].explode().unique():
                v = task_cfg.get(key)
                if isinstance(v, list | ListConfig):
                    v = [v] * len(gdf)
                gdf[c] = v
        new_logs.append(gdf)
    return pd.concat(new_logs)


def filter_task_info(logs: pd.DataFrame, keep_task_columns: list[str] | None = None) -> pd.DataFrame:
    if keep_task_columns is None:
        keep_task_columns = ["n_trials"]
    keep_task_columns = [f"task.{c}" for c in keep_task_columns]
    task_cols_to_remove = [c for c in logs.columns if c.startswith("task") and c not in keep_task_columns]
    return logs.drop(columns=task_cols_to_remove)


def maybe_convert_cost_dtype(x: Any) -> tuple[float, list[float]]:
    if isinstance(x, int | float):
        return float(x)
    elif isinstance(x, str):
        return eval(x)
    else:
        assert isinstance(x, list)
        return x


def maybe_convert_cost_to_so(x: Any) -> float:
    if isinstance(x, list):
        return np.sum(x)  # TODO replace by hypervolume or similar
    else:
        return x


def convert_mixed_types_to_str(logs: pd.DataFrame, logger=None) -> pd.DataFrame:
    mixed_type_columns = logs.select_dtypes(include=["O"]).columns
    if logger:
        logger.debug(f"Goodybe all mixed data, ruthlessly converting {mixed_type_columns} to str...")
    for c in mixed_type_columns:
        # D = logs[c]
        # logs.drop(columns=c)
        if c == "cfg_str":
            continue
        logs[c] = logs[c].map(lambda x: str(x))
        logs[c] = logs[c].astype("str")
    return logs


def load_set(paths: list[str], set_id: str = "unknown") -> tuple[pd.DataFrame, pd.DataFrame]:
    logs = []
    for p in paths:
        fn = Path(p) / "trajectory.parquet"
        if not fn.is_file():
            fn = Path(p) / "logs.parquet"
        logs.append(pd.read_parquet(fn))

    df = pd.concat(logs).reset_index(drop=True)
    df_cfg = pd.concat([pd.read_parquet(Path(p) / "logs_cfg.parquet") for p in paths]).reset_index(drop=True)
    df["set"] = set_id
    return df, df_cfg


def process_logs(logs: pd.DataFrame, keep_task_columns: list[str] | None = None) -> pd.DataFrame:
    if keep_task_columns is None:
        keep_task_columns = ["task.n_trials"]
    logger.debug("Processing raw logs. Normalize n_trials and costs. Calculate trajectory (incumbent cost).")
    # logs= logs.drop(columns=["config"])
    # Filter MO costs
    logger.debug("Remove DUMMY logs...")
    logs = logs[~logs["problem_id"].str.startswith("DUMMY")]
    logs = logs[~logs["benchmark_id"].str.startswith("DUMMY")]
    logs = logs[~logs["optimizer_id"].str.startswith("DUMMY")]

    logger.debug("Handle MO costs...")
    logs["trial_value__cost_raw"] = logs["trial_value__cost"].apply(maybe_convert_cost_dtype)
    logs["trial_value__cost"] = logs["trial_value__cost_raw"].apply(maybe_convert_cost_to_so)
    logger.debug("Determine incumbent cost...")
    logs["trial_value__cost_inc"] = logs.groupby(by=["problem_id", "optimizer_id", "seed"])[
        "trial_value__cost"
    ].transform("cummin")

    logger.debug("Maybe add task info...")
    logs = maybe_postadd_task(logs)
    if "task.n_objectives" in logs:
        logs["task.is_multiobjective"] = logs["task.n_objectives"] > 1
    logger.debug("Infer scenario...")
    logs = add_scenario_type(logs)

    # Check for scalarized MO, we want to keep the cost vector
    if "trial_value__additional_info" in logs:
        ids_mo = (logs["scenario"] == "multi-objective") & (
            logs["trial_value__additional_info"].apply(lambda x: "cost" in x)
        )
        if len(ids_mo) > 0:
            logs[ids_mo]["trial_value__cost_raw"] = logs[ids_mo]["trial_value__additional_info"].apply(
                lambda x: x["cost"]
            )

    logger.debug(f"Remove task info, only keep {keep_task_columns}...")
    logs = filter_task_info(logs, keep_task_columns)

    # Convert config to object
    logger.debug("Save config as a string to avoid mixed type columns...")
    logs["trial_info__config"] = logs["trial_info__config"].apply(lambda x: str(x))

    # Add time
    logger.debug("Calculate the elapsed time...")
    logs = logs.groupby(by=["problem_id", "optimizer_id", "seed"]).apply(calc_time).reset_index(drop=True)

    logs = convert_mixed_types_to_str(logs, logger)
    logger.debug("Done 😪🙂")
    return logs


def normalize_logs(logs: pd.DataFrame) -> pd.DataFrame:
    logger.info("Start normalization...")
    logger.info("Normalize n_trials...")
    logs["n_trials_norm"] = logs.groupby("problem_id")["n_trials"].transform(normalize)
    logger.info("Normalize cost...")
    # Handle MO
    ids_mo = logs["scenario"] == "multi-objective"
    if len(ids_mo) > 0 and "hypervolume" in logs:
        hv = logs.loc[ids_mo, "hypervolume"]
        logs.loc[ids_mo, "trial_value__cost"] = -hv  # higher is better
        logs["trial_value__cost"] = logs["trial_value__cost"].astype("float64")
        logs["trial_value__cost_inc"] = logs["trial_value__cost"].transform("cummin")
    logs["trial_value__cost_norm"] = logs.groupby("problem_id")["trial_value__cost"].transform(normalize)
    logger.info("Calc normalized incumbent cost...")
    logs["trial_value__cost_inc_norm"] = logs.groupby(by=["problem_id", "optimizer_id", "seed"])[
        "trial_value__cost_norm"
    ].transform("cummin")
    if "time" not in logs:
        logs["time"] = 0
    logger.info("Normalize time...")
    logs["time_norm"] = logs.groupby("problem_id")["time"].transform(normalize)
    logs = convert_mixed_types_to_str(logs, logger)
    logger.info("Done.")
    return logs


def calc_time(D: pd.DataFrame) -> pd.Series:
    trialtime = D["trial_value__virtual_time"]
    nulltime = D["trial_value__starttime"] - D["trial_value__starttime"].min()
    trialtime_cum = trialtime.cumsum()
    elapsed = nulltime + trialtime_cum
    elapsed.name = "time"
    D["time"] = elapsed
    return D


def normalize(S: pd.Series, epsilon: float = 1e-8) -> pd.Series:
    return (S - S.min()) / (S.max() - S.min() + epsilon)


def get_interpolated_performance_df(
    logs: pd.DataFrame,
    n_points: int = 20,
    x_column: str = "n_trials_norm",
    interpolation_columns: list[str] | None = None,
) -> pd.DataFrame:
    """Get performance dataframe for plotting.

    Interpolated at regular intervals.

    Parameters
    ----------
    logs : pd.DataFrame
        Preprocessed logs.
    n_points : int, optional
        Number of interpolation steps, by default 20
    x_column : str, optional
        The x-axis column to interpolate by, by default 'n_trials_norm'

    Raises:
    ------
    ValueError
        When x_column missing in dataframe.

    Returns:
    -------
    pd.DataFrame
        Performance data frame for plotting
    """
    if interpolation_columns is None:
        interpolation_columns = [
            "trial_value__cost",
            "trial_value__cost_norm",
            "trial_value__cost_inc",
            "trial_value__cost_inc_norm",
        ]
    logger.info("Create dataframe for neat plotting by aligning x-axis / interpolating budget.")

    if x_column not in logs:
        msg = f"x_column `{x_column}` not in logs! Did you call `carps.analysis.process_data.process_logs` on the raw logs?"
        raise ValueError(msg)

    # interpolation_columns = [
    #     c for c in logs.columns if c != x_column and c not in identifier_columns and not c.startswith("problem")]
    group_keys = ["scenario", "set", "benchmark_id", "optimizer_id", "problem_id", "seed"]
    x = np.linspace(0, 1, n_points + 1)
    D = []
    for gid, gdf in logs.groupby(by=group_keys):
        metadata = dict(zip(group_keys, gid, strict=False))
        performance_data = {}
        performance_data[x_column] = x
        for icol in interpolation_columns:
            if icol in gdf:
                xp = gdf[x_column].to_numpy()
                fp = gdf[icol].to_numpy()
                y = np.interp(x=x, xp=xp, fp=fp)
                performance_data[icol] = y
        performance_data.update(metadata)
        D.append(pd.DataFrame(performance_data))
    return pd.concat(D).reset_index(drop=True)


def load_logs(rundir: str):
    logs_fn = Path(rundir) / "logs.csv"
    logs_cfg_fn = logs_fn.parent / "logs_cfg.csv"

    logger.info(f"Load logs from `{logs_fn}` and associated configs from {logs_cfg_fn}. Preprocess logs.")

    if not logs_fn.is_file() or not logs_cfg_fn.is_file():
        msg = f"No logs found at rundir '{rundir}'. If you used the file logger, you can gather the data with `python -m carps.analysis.gather_data <rundir>`."
        raise RuntimeError(msg)

    df = pd.read_csv(logs_fn)
    df = normalize_logs(df)
    df_cfg = pd.read_csv(logs_cfg_fn)
    return df, df_cfg


# NOTE(eddiebergman): Use `n_processes=None` as default, which uses `os.cpu_count()` in `Pool`
def filelogs_to_df(
    rundir: str, log_fn: str = "trial_logs.jsonl", n_processes: int | None = None
) -> tuple[pd.DataFrame, pd.DataFrame]:
    logger.info(f"Get rundirs from {rundir}...")
    rundirs = get_run_dirs(rundir)
    logger.info(f"Found {len(rundirs)} runs. Load data...")
    partial_load_log = partial(load_log, log_fn=log_fn)
    results = map_multiprocessing(partial_load_log, rundirs, n_processes=n_processes)
    df = pd.concat(results).reset_index(drop=True)
    logger.info("Done. Do some preprocessing...")
    df_cfg = pd.DataFrame([{"cfg_fn": k, "cfg_str": v} for k, v in df["cfg_str"].unique()])
    df_cfg.loc[:, "experiment_id"] = np.arange(0, len(df_cfg))
    df["experiment_id"] = df["cfg_fn"].apply(lambda x: np.where(df_cfg["cfg_fn"].to_numpy() == x)[0][0])
    df_cfg.loc[:, "cfg_str"] = df_cfg["cfg_str"].apply(lambda x: x.replace("\n", "\\n"))
    del df["cfg_str"]
    del df["cfg_fn"]
    logger.info("Done. Saving to file...")
    # df = df.map(lambda x: x if not isinstance(x, list) else str(x))
    df.to_csv(Path(rundir) / "logs.csv", index=False)
    df_cfg.to_csv(Path(rundir) / "logs_cfg.csv", index=False)
    df = convert_mixed_types_to_str(df)
    df_cfg = convert_mixed_types_to_str(df_cfg)
    df.to_parquet(Path(rundir) / "logs.parquet", index=False)
    df_cfg.to_parquet(Path(rundir) / "logs_cfg.parquet", index=False)
    logger.info("Done. 😊")
    return df, df_cfg


if __name__ == "__main__":
    fire.Fire(filelogs_to_df)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import seaborn as sns
from autorank import create_report
from autorank._util import get_sorted_rank_groups

from carps.analysis.run_autorank import calc_critical_difference, custom_latex_table, get_df_crit
from carps.analysis.utils import savefig, get_color_palette, filter_only_final_performance

if TYPE_CHECKING:
    import pandas as pd


def plot_ranking(
    gdf: pd.DataFrame,
    scenario: str,
    set_id: str,
    perf_col: str = "trial_value__cost_inc_norm",
    problem_prefix: str = "",
) -> None:
    fpath = Path("figures/ranking")
    fpath.mkdir(exist_ok=True, parents=True)
    identifier = f"{scenario}_{set_id}"
    label = f"tab:stat_results_{identifier}"
    result = calc_critical_difference(gdf, identifier=identifier, figsize=(8, 3), perf_col=perf_col)
    # print(result)
    # try: 
    #     create_report(result)
    # except Exception as e:
    #     print(e)
    # table_str = custom_latex_table(result, label=label)
    # fn = Path("figures/critd/" + label[len("tab:") :] + ".tex")
    # fn.write_text(table_str)
    # print(table_str)
    # plt.show()

    sorted_ranks, names, groups = get_sorted_rank_groups(result, reverse=False)
    # print(sorted_ranks, names, groups)

    # # DF on normalized perf values
    # df_crit = get_df_crit(gdf, nan_handling="keep", perf_col=perf_col)
    # df_crit = df_crit.reindex(columns=names)
    # df_crit.index = [i.replace(problem_prefix + "/dev/", "") for i in df_crit.index]
    # df_crit.index = [i.replace(problem_prefix + "/test/", "") for i in df_crit.index]
    # plt.figure(figsize=(12, 12))
    # sns.heatmap(df_crit, annot=False, fmt="g", cmap="viridis_r")
    # plt.title("Performance of Optimizers per Problem (Normalized)")
    # plt.ylabel("Problem ID")
    # plt.xlabel("Optimizer")
    # savefig(plt.gcf(), fpath / f"perf_opt_per_problem_{identifier}")
    # plt.show()

    # # Df on raw values
    # # Optionally, plot the ranked data as a heatmap
    # df_crit = get_df_crit(gdf, nan_handling="keep", perf_col=perf_col)
    # df_crit = df_crit.reindex(columns=names)
    # df_crit.index = [i.replace(problem_prefix + "/dev/", "") for i in df_crit.index]
    # df_crit.index = [i.replace(problem_prefix + "/test/", "") for i in df_crit.index]
    # ranked_df = df_crit.rank(axis=1, method="min", ascending=True)

    # plt.figure(figsize=(12, 12))
    # sns.heatmap(ranked_df, annot=True, fmt="g", cmap="viridis_r")
    # plt.title("Ranking of Optimizers per Problem")
    # plt.ylabel("Problem ID")
    # plt.xlabel("Optimizer")
    # savefig(plt.gcf(), fpath / f"rank_opt_per_problem_{identifier}")
    # plt.show()

    # # Plotting the heatmap of the rank correlation matrix
    # correlation_matrix = ranked_df.corr(method="spearman")
    # plt.figure(figsize=(8, 6))
    # sns.heatmap(correlation_matrix, annot=True, cmap="coolwarm", cbar=True, square=True, fmt=".2f")
    # plt.title("Spearman Rank Correlation Matrix Between Optimizers")
    # savefig(plt.gcf(), fpath / f"spearman_rank_corr_matrix_opt_{identifier}")
    # plt.show()


    # combined
    ncols = 2
    nrows = 3
    right = 0.6
    h = 0.9
    w = 1
    factor = 30
    hspace = 0.3
    wspace = 0.7

    fig = plt.figure(layout=None, facecolor="white", figsize=(w * factor, h * factor))
    gs = fig.add_gridspec(nrows=nrows, ncols=ncols, left=0.05, right=right,
                        hspace=hspace, wspace=wspace
                        )

    # Perf per problem (normalized)
    ax0 = fig.add_subplot(gs[:, :-1])
    df_crit = get_df_crit(gdf, nan_handling="keep", perf_col=perf_col)
    df_crit = df_crit.reindex(columns=names)
    df_crit.index = [i.replace(problem_prefix + "/dev/", "") for i in df_crit.index]
    df_crit.index = [i.replace(problem_prefix + "/test/", "") for i in df_crit.index]
    ax0 = sns.heatmap(df_crit, annot=False, fmt="g", cmap="viridis_r", ax=ax0, cbar_kws={"shrink": 0.8, "aspect": 30})
    ax0.set_title("Final Performance per Problem (Normalized)")
    ax0.set_ylabel("Problem ID")
    ax0.set_xlabel("Optimizer")


    df_finalperf = filter_only_final_performance(df=gdf)
    sorter = names
    df_finalperf = df_finalperf.sort_values(by="optimizer_id", key=lambda column: column.map(lambda e: sorter.index(e)))
    palette = get_color_palette(df=gdf)
    ax1 = fig.add_subplot(gs[0, -1])
    x="n_trials_norm"
    y="trial_value__cost_inc_norm"
    x = y
    hue="optimizer_id"
    y = hue
    ax1 = sns.boxplot(
        data=df_finalperf, y=y, x=x, hue=hue, palette=palette, ax=ax1
    )
    ax1.set_title("Final Performance (Normalized)")

    ax2 = fig.add_subplot(gs[1, -1])
    ax2 = sns.violinplot(
        data=df_finalperf, y=y, x=x, hue=hue, palette=palette, ax=ax2, cut=0
    )
    ax2.set_title("Final Performance (Normalized)")

    # Spearman rank correlation
    ax3 = fig.add_subplot(gs[2, -1])
    ranked_df = df_crit.rank(axis=1, method="min", ascending=True)
    correlation_matrix = ranked_df.corr(method="spearman")
    ax3 = sns.heatmap(correlation_matrix, annot=True, cmap="coolwarm", cbar=True, square=True, fmt=".2f", ax=ax3)
    ax3.set_title("Spearman Rank Correlation Matrix\nBetween Optimizers")

    # fig.set_tight_layout(True)

    savefig(fig, fpath / f"final_per_combined_{identifier}")

    plt.show()
